"""This module contains general tools for working with SAXS data with the SAXS Assistant.
This includes dimensionless GPA calculations, and other general tools."""

import numpy as np
import matplotlib.pyplot as plt


def get_dim_GPA(I, q, rg, i0, plot=False):
    """Calculates the dimensionless GPA (Guinier Peak Analysis) from the given intensity and q values.
    Parameters
    ----------
    I : numpy.ndarray
        Intensity values.
    q : numpy.ndarray
        Scattering vector values.
    rg : float
        Radius of gyration.
    i0 : float
        Intensity at zero angle (I0).
    plot : bool, optional
        If True, plots the data and fit. Default is False.
    Returns
    -------
    x_maxs : numpy.ndarray
        x values of the maxima in the dimensionless GPA plot.
    y_maxs : numpy.ndarray
        y values of the maxima in the dimensionless GPA plot.
    """
    y = (q * rg * I) / i0
    x = (q * rg) ** 2
    index_max_plot = np.argmin(abs(x - 4))

    x_mins, x_maxs, y_mins, y_maxs = min_max_function(
        y[:index_max_plot], x[:index_max_plot]
    )

    if plot:
        plt.plot(x, y, "o", label="Data")
        try:
            plt.plot(x_maxs, y_maxs, "ro", label="Maxs")
        except (ValueError, TypeError):
            plt.title("Dim GPA  ")
            plt.ylabel("qRgI(q)/I(0)")
            plt.xlabel("$(Rgq)^2$")
            plt.xlim([0, 4])
            plt.show()
    return x_maxs, y_maxs


def get_GPA(I, q, plot=False, title=None):
    """
    Computes data for GPA plot (q² vs q·I(q)) in the region q² < 0.0012.

    Parameters:
        I (np.array): Scattering intensity values
        q (np.array): Scattering vector
        plot (bool): If True, returns a matplotlib figure
        title (str): Optional plot title

    Returns:
        If plot=False:
            dict with "x" (q²), "y" (q·I(q)), and "peak_index"
        If plot=True:
            tuple (figure, {"x": ..., "y": ..., "peak_index": ...})
    """
    q_sqr = q * q
    mask = q_sqr < 0.0012  # Setting
    q_sqr = q_sqr[mask]

    i_x_q = q * I
    i_x_q = i_x_q[: len(q_sqr)]  # Match array length
    index_max_y = np.argmax(i_x_q)

    gpa_data = {"x": q_sqr, "y": i_x_q, "peak_index": index_max_y}

    if plot:
        fig, ax = plt.subplots()
        ax.plot(q_sqr, i_x_q, "o")
        ax.set_title(title if title else "GPA Plot")
        ax.axvline(x=q_sqr[index_max_y], color="r", linestyle="--", label="Peak")
        ax.set_ylabel("q·I(q)")
        ax.set_xlabel("$q^2$")
        plt.close(fig)
        return fig, gpa_data

    return gpa_data


def min_max_function(intensity, scatter_angle):
    """
    Finds local minima and maxima in an intensity array. This function
    is used for Method 1 to determine the minima and maxima of the
    residuals of the Guinier fit. Then later applied to try to figure out how
    often the residuals cross the zero line. This is used to try to get an idea
    of biased residuals.
    Parameters:
        intensity: 1D array of intensity values
        scatter_angle: 1D array of scattering angle values corresponding to the intensity
    Returns:
        x_mins, x_maxs: q-values at minima and maxima
        y_mins, y_maxs: intensity values at minima and maxima
    """
    intensity = np.array(intensity)
    scatter_angle = np.array(scatter_angle)

    index_minima, y_mins = [], []
    index_maxima, y_maxs = [], []

    for i in range(1, len(intensity) - 1):
        left, center, right = intensity[i - 1], intensity[i], intensity[i + 1]
        if center < left and center < right:
            index_minima.append(i)
            y_mins.append(center)
        if center > left and center > right:
            index_maxima.append(i)
            y_maxs.append(center)

    if not index_minima and not index_maxima:
        return None, None, None, None

    x_mins = [scatter_angle[i] for i in index_minima]
    x_maxs = [scatter_angle[i] for i in index_maxima]

    return np.array(x_mins), np.array(x_maxs), y_mins, y_maxs


def get_kratky(q, I, plot=False, title=None):
    """
    Computes Kratky plot data (q vs q²I(q)).

    Parameters:
        q (np.array): Scattering vector.
        I (np.array): Intensity values.
        plot (bool): Whether to return a matplotlib figure.
        title (str): Title for the plot.

    Returns:
        If plot=False:
            dict with "x" and "y" data
        If plot=True:
            tuple (figure, {"x": ..., "y": ...})
    """
    Iq_2 = I * (q * q)
    kratky_data = {"x": q, "y": Iq_2}

    if plot:
        fig, ax = plt.subplots()
        ax.plot(q, Iq_2, "o")
        ax.set_xlabel("q")
        ax.set_ylabel(r"$q^2 I(q)$")
        if title:
            ax.set_title(title)
        plt.close(fig)
        return fig, kratky_data

    return kratky_data



"""This module contains tools for working Rg. Including the determination using the
custom method from script called Method 1, and evaluating guinier fits from both methods, including those
obtained from RAW."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import logging

# from .utils import helpers
from .utils.helpers import get_R2
from .features import get_dim_GPA

from .features import min_max_function


def get_guinier_from_min_max_2(I, q, error, nmin, nmax, plot=False):
    """
    V2 Computes Guinier fit and associated parameters including Rg, I0, R², and their errors.
    Similarly to BioXTAS but uses the values of Rg obtained from multiple BIFT Pr calculations
    as qmin is truncated to only consider fits within the range of possible Rgs from real space.
    This is used within a loop.
    Parameters
    ----------
    I : numpy.ndarray
        Intensity values.
    q : numpy.ndarray
        Scattering vector values.
    error : numpy.ndarray
        Error values associated with the intensity.
    nmin : int
        Minimum index for the Guinier fit.
    nmax : int
        Maximum index for the Guinier fit.
    plot : bool, optional
        If True, plots the data and fit. Default is False.
    Returns
    -------
    rg : float
        Radius of gyration.
    i0 : float
        Intensity at zero angle (I0).
    fR2 : float
        Coefficient of determination (R²) for the fit.
    guinier_dict : dict
        Dictionary containing various parameters from the Guinier fit.
    ax1, ax1a, ax2, ax2a : matplotlib.axes.Axes
        Axes objects for plotting the data and fit.
    """

    def linearFunc(x, intercept, slope):
        return intercept + slope * x

    end = nmax + 1
    x = np.array(q[nmin:end]) ** 2
    y = np.log(I[nmin:end])
    err = np.abs(error[nmin:end] / I[nmin:end])

    valid = np.isfinite(y)
    x, y, err = x[valid], y[valid], err[valid]

    a_fit, cov = curve_fit(linearFunc, x, y, sigma=err, absolute_sigma=True)
    intercept, slope = a_fit
    yfit = linearFunc(x, intercept, slope)
    fR2 = get_R2(y, yfit)

    # Compute Rg and I0
    rg = np.sqrt(-3 * slope) if slope < 0 else np.nan
    i0 = np.exp(intercept)

    # Error propagation
    try:
        slope_var = cov[1, 1]
        intercept_var = cov[0, 0]
        rg_err = (0.5 / rg) * np.sqrt(3 * slope_var) if slope < 0 else np.nan
        i0_err = i0 * np.sqrt(intercept_var)
    except Exception:
        rg_err = np.nan
        i0_err = np.nan

    # qRg bounds
    qminRg = q[nmin] * rg
    qmaxRg = q[end - 1] * rg
    qmin = q[nmin]
    qmax = q[end - 1]

    # Residual analysis
    residuals = (y - yfit) / err
    x_mins, x_maxs, y_mins, y_maxs = min_max_function(residuals, x)

    # Count x-crossings
    crossed = 0
    strt = residuals[0]
    for ii in range(1, len(residuals)):
        current = residuals[ii]
        if (strt > 0 and current < 0) or (strt < 0 and current > 0):
            crossed += 1
        strt = current

    peaks_v_cross = 0
    try:
        peaks_v_cross = crossed / (len(x_mins) + len(x_maxs))
    except ZeroDivisionError:
        pass

    GPAx_maxs, GPAy_maxs = get_dim_GPA(I, q, rg, i0)
    closest_Gx = GPAx_maxs[np.argmin(abs(GPAx_maxs - 1.5))] if len(GPAx_maxs) > 0 else 0
    closest_Gy = GPAy_maxs[np.argmin(abs(GPAx_maxs - 1.5))] if len(GPAy_maxs) > 0 else 0

    len_res = max(residuals) - min(residuals)

    if plot:
        fig, ax1 = plt.subplots()
        fig, ax1a = plt.subplots()
        ax1.plot(x, y, "o", label="Data")
        ax1a.plot(x, yfit, label="Fit $R^2$ " + str(round(fR2, 4)))

        fig, ax2 = plt.subplots()
        fig, ax2a = plt.subplots()
        ax2.plot(x, residuals, "o")
        try:
            ax2.plot(x_mins, y_mins, "og")
            ax2.plot(x_maxs, y_maxs, "oc")
        except Exception:
            pass
        ax2a.plot(x, np.zeros(len(x)), "r")
    else:
        ax1 = ax1a = ax2 = ax2a = 0

    guinier_dict = {
        "Rg": [rg],
        "Rg Err": [rg_err],
        "i0": [i0],
        "I0 Err": [i0_err],
        "fit_r2": [fR2],
        "qRgmin": [qminRg],
        "qRgmax": [qmaxRg],
        "qmin": [qmin],
        "qmax": [qmax],
        "peaks/x-cross": [peaks_v_cross],
        "GPA x peak": [closest_Gx],
        "GPA y peak": [closest_Gy],
        "Res window": [len_res],
        "nmin": nmin,
        "nmax": nmax,
    }

    return rg, i0, fR2, guinier_dict, ax1, ax1a, ax2, ax2a


def calculate_guinier_Iq(q, rg, i0):
    """
    Returns I(q) calculated from Guinier approximation:
    I(q) = I0 * exp(-Rg^2 * q^2 / 3)
    Used for evaluating the Guinier fit of both methods, based on the
    agreement of the Rg and I0 values obtained from the fit.
    Parameters
    ----------
    q : numpy.ndarray
        Scattering vector values.
    rg : float
        Radius of gyration.
    i0 : float
        Intensity at zero angle (I0).
    Returns
    -------
    numpy.ndarray
        Calculated intensity values I(q) based on the Guinier approximation.
    """
    return i0 * np.exp(-(rg**2 * q**2) / 3)


def rg_method_1(q, I, err, pr_rg_list):
    """
    Performs second-pass Guinier fits using Rg range from P(r).
    Returns best-fit DataFrame and list of all candidate fits.
    """
    # print(pr_rg_list)
    gds = []

    for i in range(30):  # Start indices
        dynamic_bound = int((len(q) - i) - 1)
        for window in range(7, dynamic_bound):
            try:
                rg, i0, r2, gd, *_ = get_guinier_from_min_max_2(
                    I, q, err, i, i + window
                )
            except Exception:
                continue

            qrgmin = rg * q[i]
            qrgmax = rg * q[i + window - 1]

            if rg >= min(pr_rg_list) * 0.91 and rg <= max(pr_rg_list) * 1.09:
                if r2 > 0.74 and qrgmax < 1.4 and qrgmin < 0.9 and qrgmax > 0.9:
                    gds.append(gd)

    if not gds:
        return None, []

    # Collect and return all good fits
    all_gds_df = pd.concat(
        [pd.DataFrame.from_dict(gd) for gd in gds], ignore_index=True
    )
    return all_gds_df, gds


def select_final_rg_from_candidates(gds, q, I, err, sample_id="unknown"):
    """
    From a list of Guinier fits (Method 1), filters and selects the best candidate.
    Returns final Guinier DataFrame or None if unresolved.
    """
    if not gds:
        logging.warning(f"No Guinier fits found for {sample_id}")
        return None

    GD = pd.DataFrame.from_dict(gds[0])
    for gd in gds[1:]:
        GD = pd.concat([GD, pd.DataFrame.from_dict(gd)], axis=0, ignore_index=True)

    filt1 = GD[GD["qRgmin"] <= 0.9]
    # print('filt1')
    # display(filt1)
    if len(filt1["GPA x peak"]) < 1:
        return None

    filt2 = filt1[filt1["qRgmax"] < 1.4]
    if len(filt2) < 1:
        filt2 = filt1[filt1["qRgmax"] < 1.5]

    filt3 = filt2[filt2["peaks/x-cross"] <= 1.33]
    if len(filt3) < 1:
        return None

    filt4 = (
        filt3.copy()
        .sort_values(by=["peaks/x-cross", "qRgmin"], ascending=[False, True])
        .reset_index(drop=True)
    )
    # print('Filt4')
    # display(filt4)
    # Bin Rgs
    filter_df = filt4.round({"Rg": 0, "qRgmin": 2})
    u_rgs = np.sort(filter_df["Rg"].unique())
    if len(u_rgs) < 1:
        return None

    # Generate bin edges (10% increments)
    frst_l = [u_rgs[0]]
    while frst_l[-1] < u_rgs[-1]:
        frst_l.append(round(frst_l[-1] * 1.10) + 1)
    if len(frst_l) < 2:
        frst_l.append(frst_l[-1] + 1)

    # Histogram
    h_vals, bin_edges = np.histogram(filter_df["Rg"], bins=frst_l)
    try:
        most_occur_idx = np.argmax(h_vals)
        rg_lb = bin_edges[most_occur_idx]
        rg_ub = bin_edges[most_occur_idx + 1]
    except Exception:
        logging.warning(f"Histogram selection failed for {sample_id}")
        return None

    f1 = filter_df[(filter_df["Rg"] >= rg_lb) & (filter_df["Rg"] < rg_ub)].sort_values(
        by=["nmin", "nmax"]
    )
    if len(f1) < 1:
        return None

    nmin_new = int(f1["nmin"].min())
    nmax_new = int(f1["nmax"].max())

    # New round of Guinier fitting centered around this Rg window
    new_gds = []
    range_expand = int((nmax_new - nmin_new) * 0.15)
    for i in range(range_expand):
        for jj in range(-range_expand, range_expand):
            try:
                rg, i0, r2, gd, *_ = get_guinier_from_min_max_2(
                    I, q, err, nmin_new + i, nmax_new - jj, plot=False
                )
                if r2 > 0.73:
                    new_gds.append(gd)
            except:
                continue

    if not new_gds:
        return None

    final_df = pd.DataFrame.from_dict(new_gds[0])
    for gd in new_gds[1:]:
        final_df = pd.concat(
            [final_df, pd.DataFrame.from_dict(gd)], axis=0, ignore_index=True
        )

    # Apply quality filters
    filt1 = final_df[final_df["qRgmin"] <= 0.9]
    if len(filt1) < 1:
        # print('returned non lst -1')
        return None

    filt2 = filt1[filt1["qRgmax"] < 1.4]
    if len(filt2) < 1:
        filt2 = filt1[filt1["qRgmax"] < 1.55]

    filt3 = filt2[filt2["peaks/x-cross"] <= 1.33]
    if len(filt3) < 1:
        # print('Returned non last')
        return None

    # Final sort and return
    filt4 = filt3.sort_values(
        by=["peaks/x-cross", "qRgmin"], ascending=[False, True]
    ).reset_index(drop=True)
    # print('-------------------Final ---------------------------------------')
    # display(filt4)
    return filt4


def select_best_rg_method(
    q,
    I,
    err,
    rg_auto,
    i0_auto,
    rg1,
    i01,
    rg_auto_err=None,
    i0_auto_err=None,
    rg1_err=None,
    i01_err=None,
    r2_auto=None,
    r2_1=None,
    sample_id="unknown",
):
    """
    Compares AutoRg and Method 1 fits using residuals and selects the best.
    Returns: final_rg, final_i0, selected_method
    """
    try:
        # Generate Guinier I(q) predictions
        I_fit_auto = calculate_guinier_Iq(q, rg_auto, i0_auto)
        I_fit_1 = calculate_guinier_Iq(q, rg1, i01)

        # Compute residual stats
        res_auto = compute_guinier_residuals_Limited(I, I_fit_auto, err, q, rg_auto)
        res_1 = compute_guinier_residuals_Limited(I, I_fit_1, err, q, rg1)

        mean_auto = res_auto["residual_mean"][0]
        mean_1 = res_1["residual_mean"][0]

        if mean_auto <= mean_1:
            selected = "AutoRg"
            final_rg, final_i0 = rg_auto, i0_auto
            final_rg_err, final_i0_err = rg_auto_err, i0_auto_err
            final_r2 = r2_auto
            final_res = res_auto
            final_qrgmin = rg_auto * min(q)
            final_qrgmax = rg_auto * max(q)

        else:
            selected = "Method 1"
            final_rg, final_i0 = rg1, i01
            final_rg_err, final_i0_err = rg1_err, i01_err
            final_r2 = r2_1
            final_res = res_1
            final_qrgmin = rg1 * min(q)
            final_qrgmax = rg1 * max(q)

        return {
            "Final Rg": final_rg,
            "Final I0": final_i0,
            "Final Rg Err": final_rg_err,
            "Final I0 Err": final_i0_err,
            "Final R²": final_r2,
            "Selected Method": selected,
            "Residual Stats": final_res,
            "Final qRg min": final_qrgmin,
            "Final qRg max": final_qrgmax,
        }

    except Exception as e:
        logging.warning(f"Failed to select best Rg method for {sample_id}: {e}")
        return None


def compute_guinier_residuals_Limited(y_exp, y_fit, y_err, q, rg, num_params=1):
    valid_mask = ~np.isnan(y_exp) & ~np.isnan(y_fit)
    # Apply the mask
    qrg = q[valid_mask] * rg
    y_exp = y_exp[valid_mask][qrg <= 2]
    y_fit = y_fit[valid_mask][qrg <= 2]
    y_err = np.array(y_err)[valid_mask][qrg <= 2] if y_err is not None else None
    residuals = y_exp - y_fit

    if y_err is not None:
        chi_squared = np.sum((residuals / y_err) ** 2)
    else:
        chi_squared = np.sum(residuals**2)

    reduced_chi_squared = (
        chi_squared / (len(residuals) - num_params)
        if len(residuals) > num_params
        else np.nan
    )

    return {
        "residual_mean": [np.mean(np.abs(residuals))],
        "residual_median": [np.median(np.abs(residuals))],
        "residual_std": [np.std(residuals)],
        "residual_max": [np.max(np.abs(residuals))],
        "Reduced Chi Squared": [reduced_chi_squared],
        # 'residual_array': residuals
    }


def get_residuals_guinier_plot(I, q, error, nmin, nmax, extra=2):
    def linearFunc(x, intercept, slope):
        return intercept + slope * x

    # nmin_ext = max(0, nmin - extra)
    nmin_ext = 0

    # nmax_ext = min(len(q), nmax + extra + 1)
    nmax_ext = len(q) - 1

    x_fit = np.array(q[nmin : nmax + 1]) ** 2
    y_fit = np.log(I[nmin : nmax + 1])
    err_fit = np.abs(error[nmin : nmax + 1] / I[nmin : nmax + 1])
    print("just did first log")

    valid = np.isfinite(y_fit)
    x_fit, y_fit, err_fit = x_fit[valid], y_fit[valid], err_fit[valid]

    a_fit, cov = curve_fit(linearFunc, x_fit, y_fit, sigma=err_fit, absolute_sigma=True)
    intercept, slope = a_fit
    y_model = linearFunc(x_fit, intercept, slope)

    residuals = (y_fit - y_model) / err_fit
    x_mins, x_maxs, y_mins, y_maxs = min_max_function(residuals, x_fit)

    x_all = np.array(q[nmin_ext:nmax_ext]) ** 2
    with np.errstate(divide="ignore", invalid="ignore"):
        y_all = np.log(I[nmin_ext:nmax_ext])
    err_all = np.abs(error[nmin_ext:nmax_ext] / I[nmin_ext:nmax_ext])
    valid = np.isfinite(y_all)
    x_all, y_all, err_all = x_all[valid], y_all[valid], err_all[valid]
    print("did second log")

    return {
        "x_all": x_all,  # This all the q**2 values that passed mask
        "y_all": y_all,  # All Ln(i(q)) that arent nan
        "x_fit": x_fit,  # The actual X_values of the fit (q**2)
        "y_fit": y_fit,  # The actual Ln(I(q)) values NOT the fit but the data used for it
        "residuals": residuals,  # The resdiauls in the fit
        "x_mins": x_mins,  # This if want to plot the mins and max
        "x_maxs": x_maxs,
        "y_mins": y_mins,
        "y_maxs": y_maxs,
        "nmin_ext": nmin_ext,  # This is always just the first q value
        "nmax_ext": nmax_ext,  # Always the last q value
        "y_model": y_model,  # This is the actual fit model, the straight line
        "Rg": (-3 * slope) ** (1 / 2),
        "i0": intercept,
    }



"""This module contains helper functions for data processing in the Gormley Lab.
It includes a function to truncate SAXS data, check if files are in a folder"""

import os
from IPython.display import Javascript, display

import shutil
import numpy as np


def SAXS_truncator(input_folder, output_folder, truncation_value=0.251):
    """
    This take the path of a folder containing data not truncated (input folder) and then this will truncate the files to q<0.251.
    This ignores and removes text lines that start with "#", and only keeps the data rows where the first column (q values) is less than the truncation value (default is 0.251).
    and be saved to the path of a folder (the output folder). The new files will be saved with the current naming convention used
    at the Gormley Lab. Ex: a untruncated data file named "Polymer_5_Den_s.dat" will be saved as "Polymer_5_Den_s_SAXS.dat"

    Note:
    If the initial imports for Cesar_SAXS are ran no further imports needed, otherwise import (1) os and (2) pandas as pd
    Parameters:
    ------------
    input_folder: path, str
      The path of the folder containing the files to be truncated, this is passed as a string
    output_folder: path, str
      The path of the folder to store the truncated files, this is passed as a string

    """

    # loop over files in the input folder
    for filename in os.listdir(input_folder):
        if filename.endswith(".dat"):
            # open the input file
            with open(os.path.join(input_folder, filename), "r") as f:
                # create a list to hold the rows of data
                data_rows = []
                # loop over lines in the file
                for line in f:
                    # skip any lines that start with "#"
                    if not line.startswith("#"):
                        # split the line into three columns
                        columns = line.strip().split()
                        # check if the second column is less than 0.251
                        if float(columns[0]) < truncation_value:
                            # add the row to the list of data rows
                            data_rows.append(line)
            # create the output file
            with open(
                os.path.join(
                    output_folder,
                    filename.split(".")[0] + "_SAXS." + filename.split(".")[1],
                ),
                "w",
            ) as f:
                # write the data rows to the output file
                for row in data_rows:
                    f.write(row)


def merge_dicts(*dicts):
    """
    Merges any number of dictionaries. If keys overlap, the last one passed takes precedence.
    Useful for combining dictionaries in an analysis folder, where you might have
    multiple dictionaries with the same keys or ran execution on multiple instances.
    Parameters:
    -----------
    *dicts : dict
        Any number of dictionaries to merge.
    """
    merged = {}
    for d in dicts:
        if isinstance(d, dict):
            merged.update(d)
        else:
            raise TypeError(f"Expected dict, got {type(d)}")
    return merged


def check_files_in_directory(file_series, directory):
    """
    Check if files listed in a pandas Series are present in a specified directory.
    Parameters:
    -----------
    file_series : pd.Series
        A pandas Series containing file names to check.
    directory : str
        The path to the directory where the files should be checked.
    """

    # List all files in the given directory
    directory_files = set(os.listdir(directory))

    # Initialize an empty list to store missing files
    missing_files = []

    # Iterate over each file name in the pandas Series
    for file_name in file_series:
        if file_name not in directory_files:
            missing_files.append(file_name)

    # Print missing files, if any
    if missing_files:
        print("Missing files:", str(len(missing_files)))
        for missing_file in missing_files:
            print(missing_file)
    else:
        print("All files are present.")


def copy_files(file_list, source_dir, destination_dir):
    """
    Copies specified files from the source directory to a new destination directory.
    Used for checking files code did bad on

    Parameters:
    -----------
    - file_list: List of file names to be copied.
    - source_dir: Directory where the original files are stored.
    - destination_dir: Directory where the files will be copied.

    The function ensures that the original files are not modified.
    """

    # Ensure destination directory exists
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)  # Create the directory if it doesn’t exist
        print(f"Created destination directory: {destination_dir}")

    # Copy each file
    for file_name in file_list:
        source_path = os.path.join(source_dir, file_name)
        dest_path = os.path.join(destination_dir, file_name)

        if os.path.exists(source_path):  # Check if the file exists before copying
            shutil.copy2(source_path, dest_path)  # Copy with metadata preserved
            print(f"Copied: {file_name} → {destination_dir}")
        else:
            print(f"⚠️ Warning: {file_name} not found in {source_dir}")

    print("Copy process completed!")


def beep():
    """
    Used for notifications
    """
    display(
        Javascript("""
      (async () => {
          const context = new AudioContext();
          const o = context.createOscillator();
          const g = context.createGain();
          o.type = "sine";
          o.connect(g);
          g.connect(context.destination);
          o.start();
          g.gain.exponentialRampToValueAtTime(0.00001, context.currentTime + 1);
      })();
  """)
    )


def get_R2(true_y, fitted_y):
    """
    Gets the fit R^2 value using the formula
    R^2 = 1 - (Sum of squares of residuals / Total sum of squares)

    Parameters
    ----------
    true_y: array
      An array/ series of the original depended variable data
    fitted_y: array
      An array/ series of the fitted dependent variable
    """
    residuals = true_y - fitted_y
    sum_res = np.sum(residuals**2)
    squares_sum = np.sum((true_y - np.mean(true_y)) ** 2)
    return 1 - (sum_res / squares_sum)


def reprocess_sasbdb_q_values(sasbdb_q, sasbdb_I, sasbdb_error):
    """
    Reprocess SASBDB q-values to match lab q-spacing.
    This function is applied to all data used in SAXS Assistant to ensure consistency in q-values.
    It selects q-values from SASBDB/Other data that fall within the range of q-values used in your lab.
    Originally, made to process data from the SASBDB, but can be used for any dataset with q-values.
    It finds the closest q-values from SASBDB to the lab q-values, avoiding duplicates.

    Parameters:
        sasbdb_q (numpy array): q-values from SASBDB.
        sasbdb_I (numpy array): Intensity values from SASBDB.
        sasbdb_error (numpy array or None): Error values from SASBDB (can be None).
        lab_q_array (numpy array): q-values used in your lab.

    Returns:
        dict: Dictionary with selected q, I(q), and errors (if available).
    """
    lab_q_array = np.array(
        [
            0.005,
            0.006,
            0.007,
            0.008,
            0.009,
            0.01,
            0.011,
            0.012,
            0.013,
            0.014,
            0.015,
            0.016,
            0.017,
            0.018,
            0.019,
            0.02,
            0.021,
            0.022,
            0.023,
            0.024,
            0.025,
            0.026,
            0.027,
            0.028,
            0.029,
            0.03,
            0.031,
            0.032,
            0.033,
            0.034,
            0.035,
            0.036,
            0.037,
            0.038,
            0.039,
            0.04,
            0.041,
            0.042,
            0.043,
            0.044,
            0.045,
            0.046,
            0.047,
            0.048,
            0.049,
            0.05,
            0.052,
            0.054,
            0.056,
            0.058,
            0.06,
            0.062,
            0.064,
            0.066,
            0.068,
            0.07,
            0.072,
            0.074,
            0.076,
            0.078,
            0.08,
            0.082,
            0.084,
            0.086,
            0.088,
            0.09,
            0.092,
            0.094,
            0.096,
            0.098,
            0.1,
            0.105,
            0.11,
            0.115,
            0.12,
            0.125,
            0.13,
            0.135,
            0.14,
            0.145,
            0.15,
            0.155,
            0.16,
            0.165,
            0.17,
            0.175,
            0.18,
            0.185,
            0.19,
            0.195,
            0.2,
            0.205,
            0.21,
            0.215,
            0.22,
            0.225,
            0.23,
            0.235,
            0.24,
            0.245,
            0.25,
            0.255,
            0.26,
            0.265,
            0.27,
            0.275,
            0.28,
            0.285,
            0.29,
            0.295,
            0.3,
            0.305,
            0.31,
            0.315,
            0.32,
            0.325,
            0.33,
            0.335,
            0.34,
            0.345,
            0.35,
            0.355,
            0.36,
            0.365,
            0.37,
            0.375,
            0.38,
            0.385,
            0.39,
            0.395,
            0.4,
            0.405,
            0.41,
            0.415,
            0.42,
            0.425,
            0.43,
            0.435,
            0.44,
            0.445,
            0.45,
            0.455,
            0.46,
            0.465,
            0.47,
            0.475,
            0.48,
            0.485,
            0.49,
            0.495,
            0.5,
            0.51,
            0.52,
            0.53,
            0.54,
            0.55,
            0.56,
            0.57,
            0.58,
            0.59,
            0.6,
            0.61,
            0.62,
            0.63,
            0.64,
            0.65,
            0.66,
            0.67,
            0.68,
            0.69,
            0.7,
            0.71,
            0.72,
            0.73,
            0.74,
            0.75,
            0.76,
            0.77,
            0.78,
            0.79,
            0.8,
            0.81,
            0.82,
            0.83,
            0.84,
            0.85,
            0.86,
            0.87,
            0.88,
            0.89,
            0.9,
            0.91,
            0.92,
            0.93,
            0.94,
            0.95,
            0.96,
            0.97,
            0.98,
            0.99,
            1.0,
            1.03,
            1.06,
            1.09,
            1.12,
            1.15,
            1.18,
            1.21,
            1.24,
            1.27,
            1.3,
            1.33,
            1.36,
            1.39,
            1.42,
            1.45,
            1.48,
            1.51,
            1.54,
            1.57,
            1.6,
            1.63,
            1.66,
            1.69,
            1.72,
            1.75,
            1.78,
            1.81,
            1.84,
            1.87,
            1.9,
            1.93,
            1.96,
            1.99,
            2.02,
            2.05,
            2.08,
            2.11,
            2.14,
            2.17,
            2.2,
            2.23,
            2.26,
            2.29,
            2.32,
            2.35,
            2.38,
            2.41,
            2.44,
            2.47,
            2.5,
            2.53,
            2.56,
            2.59,
            2.62,
            2.65,
            2.68,
            2.71,
            2.74,
            2.77,
            2.8,
            2.83,
            2.86,
            2.89,
            2.92,
            2.95,
            2.98,
            3.01,
            3.04,
            3.07,
            3.1,
            3.13,
            3.16,
            3.19,
        ]
    )
    # Step 1: Get min/max q-values from SASBDB
    q_min_sasbdb, q_max_sasbdb = np.min(sasbdb_q), np.max(sasbdb_q)

    # Step 2: Find first and last matching q-values in lab data
    q_min_match = lab_q_array[lab_q_array >= q_min_sasbdb].min()
    q_max_match = lab_q_array[lab_q_array <= q_max_sasbdb].max()

    # Get the q-values from lab that fall within the valid range
    valid_lab_qs = lab_q_array[
        (lab_q_array >= q_min_match) & (lab_q_array <= q_max_match)
    ]

    # Step 3: Find closest q-values from SASBDB to lab q-values
    selected_q, selected_I, selected_error = [], [], []

    for q_target in valid_lab_qs:
        # Get the index of the closest SASBDB q-value to the lab q-value
        idx_closest = np.argmin(np.abs(sasbdb_q - q_target))

        # Avoid duplicate selections
        if selected_q and sasbdb_q[idx_closest] == selected_q[-1]:
            continue  # Skip duplicates

        # Store selected q, I(q), and error (if available)
        selected_q.append(sasbdb_q[idx_closest])
        selected_I.append(sasbdb_I[idx_closest])
        selected_error.append(
            sasbdb_error[idx_closest] if sasbdb_error is not None else None
        )

    # Convert to numpy arrays
    selected_q = np.array(selected_q)
    selected_I = np.array(selected_I)
    selected_error = np.array(selected_error) if sasbdb_error is not None else None

    # Return reprocessed dataset
    return {"q": selected_q, "I": selected_I, "error": selected_error}



# -*- coding: utf-8 -*-
"""SASCalc

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/16WemERFWTY_7Jb3FvuQf1bXT19B8-egU
"""

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import object, range, map, zip

# from io import open
"""
Created on December 12, 2015

@author: Jesse B. Hopkins

#******************************************************************************
# This file is part of RAW.
#
#    RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with RAW.  If not, see <http://www.gnu.org/licenses/>.
#
#******************************************************************************

The purpose of this module is to contain functions for calculating
values from SAXS profiles. These are intended to be automated
functions, including calculation of rg and molecular weight.

It also contains functions for calling outside packages for use in RAW, like DAMMIF.
"""
try:
    import queue
except Exception:
    import Queue as queue

# import os
import time
import subprocess
import threading
import platform
import re
import math
import traceback
import copy
import tempfile

import numpy as np
from scipy import integrate as integrate
import scipy.interpolate
import scipy.signal
import scipy.stats as stats

# from scipy.constants import Avogadro
from numba import jit

# import sasexceptions as SASExceptions
from . import sasexceptions as SASExceptions

# import bioxtasraw.SASExceptions as SASExceptions
# import bioxtasraw.SASM as SASM
from . import sasm as SASM


# Define the rg fit function
@jit(nopython=True, cache=True, parallel=False)
def linear_func(x, a, b):
    return a + b * x


@jit(nopython=True, cache=True, parallel=False)
def weighted_lin_reg(x, y, err):
    weights = 1.0 / (err) ** 2.0

    w_sum = weights.sum()
    wy_sum = (weights * y).sum()
    wx_sum = (weights * x).sum()
    wxsq_sum = (weights * x**2.0).sum()
    wxy_sum = (weights * x * y).sum()

    delta = weights.sum() * wxsq_sum - (wx_sum) ** 2.0

    if delta != 0:
        a = (wxsq_sum * wy_sum - wx_sum * wxy_sum) / delta
        b = (w_sum * wxy_sum - wx_sum * wy_sum) / delta

        cov_a = wxsq_sum / delta
        cov_b = w_sum / delta
    else:
        a = -1
        b = -1
        cov_a = -1
        cov_b = -1

    return a, b, cov_a, cov_b


@jit(nopython=True, cache=True, parallel=False)
def lin_reg(x, y):
    x_sum = x.sum()
    xsq_sum = (x**2).sum()
    y_sum = y.sum()
    xy_sum = (x * y).sum()
    n = len(x)

    delta = n * xsq_sum - x_sum**2.0

    if delta != 0:
        a = (xsq_sum * y_sum - x_sum * xy_sum) / delta
        b = (n * xy_sum - x_sum * y_sum) / delta

        cov_y = (1.0 / (n - 2.0)) * ((y - a - b * x) ** 2.0).sum()
        cov_a = cov_y * (xsq_sum / delta)
        cov_b = cov_y * (n / delta)
    else:
        a = -1
        b = -1
        cov_a = -1
        cov_b = -1

    return a, b, cov_a, cov_b


@jit(nopython=True, cache=True, parallel=False)
def calcRg(q, i, err, transform=True, error_weight=True):
    if transform:
        # Start out by transforming as usual.
        x = np.square(q)
        y = np.log(i)
        yerr = np.absolute(
            err / i
        )  # I know it looks odd, but it's correct for a natural log
        x = x[np.where(np.isfinite(y))]
        yerr = yerr[np.where(np.isfinite(y))]
        y = y[np.where(np.isfinite(y))]
    else:
        x = q
        y = i
        yerr = err

    if error_weight:
        if np.any(yerr == 0):
            error_weight = False

    if error_weight:
        a, b, cov_a, cov_b = weighted_lin_reg(x, y, yerr)
    else:
        a, b, cov_a, cov_b = lin_reg(x, y)

    if b < 0:
        RG = np.sqrt(-3.0 * b)
        I0 = np.exp(a)

        # error in rg and i0 is calculated by noting that q(x)+/-Dq has Dq=abs(dq/dx)Dx, where q(x) is your function you're using
        # on the quantity x+/-Dx, with Dq and Dx as the uncertainties and dq/dx the derviative of q with respect to x.
        RGer = np.absolute(0.5 * (np.sqrt(-3.0 / b))) * np.sqrt(np.absolute(cov_b))
        I0er = I0 * np.sqrt(np.absolute(cov_a))

    else:
        RG = -1
        I0 = -1
        RGer = -1
        I0er = -1

    return RG, I0, RGer, I0er, a, b


def estimate_guinier_error(q, i, err, transform=True, error_weight=True):
    if transform:
        # Start out by transforming as usual.
        x = np.square(q)
        y = np.log(i)
        yerr = np.absolute(
            err / i
        )  # I know it looks odd, but it's correct for a natural log
    else:
        x = q
        y = i
        yerr = err

    win_size = len(x)

    if win_size < 10:
        est_rg_err = None
        est_i0_err = None
    else:
        var = win_size // 10
        if var > 12:
            step = int(np.ceil(var / 12.0))
        else:
            step = 1
        rg_list = []
        i0_list = []

        for li in range(0, var + 1, step):
            for ri in range(0, var + 1, step):
                if ri == 0:
                    Rg, I0, Rger, I0er, a, b = calcRg(
                        x[li:],
                        y[li:],
                        yerr[li:],
                        transform=transform,
                        error_weight=error_weight,
                    )
                else:
                    Rg, I0, Rger, I0er, a, b = calcRg(
                        x[li:-ri],
                        y[li:-ri],
                        yerr[li:-ri],
                        transform=transform,
                        error_weight=error_weight,
                    )

                rg_list.append(Rg)
                i0_list.append(I0)

        est_rg_err = np.array(rg_list).std()
        est_i0_err = np.array(i0_list).std()

    return est_rg_err, est_i0_err


##Porod Stuff####### uncommented 01/10/2023


def calcRefMW(i0, conc, ref_i0, ref_conc, ref_mw):
    if ref_mw > 0 and ref_i0 > 0 and ref_conc > 0 and conc > 0 and i0 > 0:
        mw = (i0 * (ref_mw / (ref_i0 / ref_conc))) / conc
    else:
        mw = -1

    return mw


def vpA(q_max):
    A = (
        -2.114 * 10**6 * q_max**4
        + 2.920 * 10**6 * q_max**3
        - 1.472 * 10**6 * q_max**2
        + 3.349 * 10**5 * q_max
        - 3.577 * 10**4
    )
    return A


def vpB(q_max):
    B = 12.09 * q_max**3 - 9.39 * q_max**2 + 3.03 * q_max + 0.29
    return B


def calcVqmax(q, i, rg, i0, choice="8/Rg", qmax=None):
    vpqmax = None

    if choice == "Default":
        if rg != 0:
            vpqmax = 8.0 / rg

            if vpqmax > 0.5 or vpqmax < 0.1:
                iratio = np.abs(np.log10(i0 / i) - 2.25)
                idx = np.argmin(iratio)

                vpqmax = q[idx]

            if vpqmax > 0.5:
                vpqmax = 0.5
            elif vpqmax < 0.1:
                vpqmax = 0.1

    elif choice == "8/Rg":
        if rg != 0:
            vpqmax = 8.0 / rg

            if vpqmax > 0.5:
                vpqmax = 0.5
            elif vpqmax < 0.1:
                vpqmax = 0.1

    elif choice == "log(I0/I(q))":
        if i0 != 0:
            iratio = np.abs(np.log10(i0 / i) - 2.25)
            idx = np.argmin(iratio)

            vpqmax = q[idx]

            if vpqmax > 0.5:
                vpqmax = 0.5
            elif vpqmax < 0.1:
                vpqmax = 0.1

    elif choice == "Manual":
        vpqmax = qmax

    if vpqmax is None:
        vpqmax = min(q[-1], 0.5)

    else:
        if vpqmax > q[-1]:
            vpqmax = q[-1]
        elif vpqmax < q[0]:
            vpqmax = q[0]
        else:
            idx = np.argmin(np.abs(q - vpqmax))
            vpqmax = q[idx]

    return vpqmax


def calcVpMW(q, i, err, rg, i0, rg_qmin, vp_density, qmax):
    # These functions are used to correct the porod volume for the length of the q vector

    if qmax not in q:
        idx = np.argmin(np.abs(q - qmax))
        qmax = q[idx]
    else:
        idx = np.argwhere(q == qmax)[0][0]

    q = q[: idx + 1]
    i = i[: idx + 1]
    err = err[: idx + 1]

    if q[-1] <= 0.5 and q[-1] >= 0.1:
        A = vpA(q[-1])
        B = vpB(q[-1])
    else:
        A = 0
        B = 1

    if i0 > 0:
        # Calculate the Porod Volume
        pVolume = porodVolume(q, i, err, rg, i0, interp=True, rg_qmin=rg_qmin)

        if pVolume == -1:
            mw = -1
            pv_cor = -1

        else:
            # Correct for the length of the q vector
            pv_cor = A + B * pVolume

            mw = pv_cor * vp_density

    else:
        mw = -1
        pVolume = -1
        pv_cor = -1

    return mw, pVolume, pv_cor


def calcAbsMW(i0, conc, rho_Mprot, rho_solv, nu_bar, r0):
    d_rho = (rho_Mprot - (rho_solv * nu_bar)) * r0
    mw = (Avogadro * i0 / conc) / np.square(d_rho)
    return mw


def volumeOfCorrelation(q, i, i0):
    """Calculates the volume of correlation as the ratio of i0 to $\int q*I dq$"""
    tot = integrate.trapz(q * i, q)
    vc = i0 / tot
    return vc


def porodInvariant(q, i, start=0, stop=-1):
    return integrate.trapz(i[start:stop] * np.square(q[start:stop]), q[start:stop])


def porodVolume(q, i, err, rg, i0, start=0, stop=-1, interp=True, rg_qmin=0):
    if interp and q[0] != 0:

        def f(x):
            return i0 * np.exp((-1.0 / 3.0) * np.square(rg) * np.square(x))

        if rg_qmin > 0:
            findClosest = lambda a, l: min(l, key=lambda x: abs(x - a))
            closest_qmin = findClosest(rg_qmin, q)

            idx_min = np.where(q == closest_qmin)[0][0]

            q = q[idx_min:]
            i = i[idx_min:]
            err = err[idx_min:]

        if len(q) != 1:
            q_interp = np.arange(0, q[0], q[1] - q[0])
            i_interp = f(q_interp)
            err_interp = np.sqrt(i_interp)

            q = np.concatenate((q_interp, q))
            i = np.concatenate((i_interp, i))
            err = np.concatenate((err_interp, err))

    if len(q) != 1:
        pInvar = porodInvariant(q, i, start, stop)

        pVolume = 2 * np.pi**2 * i0 / pInvar

    else:
        pVolume = -1

    return pVolume


# Uncomented 01/10/2023 (bottom)


def autoRg(sasm, single_fit=False, error_weight=True):
    # This function automatically calculates the radius of gyration and scattering intensity at zero angle
    # from a given scattering profile. It roughly follows the method used by the autorg function in the atsas package

    q = sasm.getQ()
    i = sasm.getI()
    err = sasm.getErr()

    qmin = 0
    # 1
    try:
        rg, rger, i0, i0er, idx_min, idx_max = autoRg_inner(
            q,
            i,
            err,
            qmin,
            single_fit,
            error_weight,
            min_window=10,
            min_qrg=1.0,
            max_qrg=1.35,
            quality_thresh=0.6,
            data_range_scale=0,
            corr_coefht=2.0,
            win_length_weight=1.0,
        )
    except Exception:  # Catches unexpected numba errors, I hope
        # ** Interesting, guess just if anything in the exceptions class is called (maybe lin fit can give wrong math easy
        # e.g division by zero? Revisit to understand why this works and how well does it)
        traceback.print_exc()
        rg = -1
        rger = -1
        i0 = -1
        i0er = -1
        idx_min = -1
        idx_max = -1

    if rg == -1:
        # If we don't find a fit, relax the criteria
        try:
            rg, rger, i0, i0er, idx_min, idx_max = autoRg_inner(
                q,
                i,
                err,
                qmin,
                single_fit,
                error_weight,
                min_window=5,
                min_qrg=1.0,
                max_qrg=1.35,
                quality_thresh=0.5,
                data_range_scale=0,
                corr_coefht=2.0,
                win_length_weight=1.0,
            )
        except Exception:  # Catches unexpected numba errors, I hope
            traceback.print_exc()
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1

    if rg == -1:
        # If we don't find a fit, relax the criteria
        try:
            rg, rger, i0, i0er, idx_min, idx_max = autoRg_inner(
                q,
                i,
                err,
                qmin,
                single_fit,
                error_weight,
                min_window=10,
                min_qrg=1.0,
                max_qrg=1.35,
                quality_thresh=0.6,
                data_range_scale=100,
                corr_coefht=2.0,
                win_length_weight=1.0,
            )
        except Exception:  # Catches unexpected numba errors, I hope
            traceback.print_exc()
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1

    if rg == -1:
        # If we don't find a fit, relax the criteria
        try:
            rg, rger, i0, i0er, idx_min, idx_max = autoRg_inner(
                q,
                i,
                err,
                qmin,
                single_fit,
                error_weight,
                min_window=10,
                min_qrg=1.2,
                max_qrg=1.5,
                quality_thresh=0.3,
                data_range_scale=100,
                corr_coefht=2.0,
                win_length_weight=1.0,
            )
        except Exception:  # Catches unexpected numba errors, I hope
            traceback.print_exc()
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1

    if rg == -1:
        # If we don't find a fit, relax the criteria
        try:
            rg, rger, i0, i0er, idx_min, idx_max = autoRg_inner(
                q,
                i,
                err,
                qmin,
                single_fit,
                error_weight,
                min_window=5,
                min_qrg=1.2,
                max_qrg=1.5,
                quality_thresh=0.3,
                data_range_scale=100,
                corr_coefht=2.0,
                win_length_weight=1.0,
            )
        except Exception:  # Catches unexpected numba errors, I hope
            traceback.print_exc()
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1

    return rg, rger, i0, i0er, idx_min, idx_max


@jit(nopython=True, cache=True, parallel=False)
def autoRg_inner(
    q,
    i,
    err,
    qmin,
    single_fit,
    error_weight,
    min_window=10,
    min_qrg=1.0,
    max_qrg=1.35,
    quality_thresh=0.6,
    data_range_scale=0,
    corr_coefht=2.0,
    win_length_weight=1.0,
):
    # Pick the start of the RG fitting range. Note that in autorg, this is done
    # by looking for strong deviations at low q from aggregation or structure factor
    # or instrumental scattering, and ignoring those. This function isn't that advanced
    # so we start at 0.

    # Note, in order to speed this up using numba, I had to do some unpythonic things
    # with declaring lists ahead of time, and making sure lists didn't have multiple
    # object types in them. It makes the code a bit more messy than the original
    # version, but numba provides a significant speedup.

    # Have to pick the right Starting range to avoid various weirdnesses in the data
    data_start = (i > 0).argmax()  # **Remember first I value is highest usually

    if len(i) > 20:
        while i[data_start] < np.mean(i[-1 * int(len(i) / 20) :]) and data_start < len(
            i
        ):
            data_start = (i[data_start + 1 :] > 0).argmax() + data_start + 1

        while i[data_start] > np.mean(i[data_start : data_start + 20]) * 10:
            data_start = (i[data_start + 1 :] > 0).argmax() + data_start + 1

    # Turns out to be pretty important to pick a good initial q range for the search
    # This is just kind of determined by what looks reasonable
    if data_range_scale == 0:
        total_int_range = np.abs(
            np.mean(i[data_start : data_start + 20]) / np.mean(i[-20:])
        )
        if total_int_range < 20:
            data_range_scale = 2.5

        elif total_int_range < 100:
            data_range_scale = 5

        elif total_int_range < 1000:
            data_range_scale = 10

        else:
            data_range_scale = 100

    ##Following the atsas package, the end point of our search space is the q value
    # where the intensity has droped by an order of magnitude from the initial value.
    data_end = np.abs(i[i > 0] - i[data_start] / data_range_scale).argmin()

    i_val = i[i > 0][data_end]
    data_end = np.argwhere(i == i_val)[0][0]

    # This makes sure we're not getting some weird fluke at the end of the scattering profile.
    if data_end > len(i) / 2.0:
        found = False
        if len(i) > data_start + 20 and data_start + 20 < len(i) / 2.0:
            idx = data_start + 20
        else:
            idx = data_start
        while not found:
            idx = idx + 1
            if i[idx] < i[data_start] / data_range_scale:
                found = True
            elif idx == len(q) - 1:
                found = True
        data_end = idx

    # Start out by transforming as usual.
    qs = np.square(q)
    il = np.log(i)
    iler = np.absolute(err / i)

    # Pick a minimum fitting window size. 10 is consistent with atsas autorg.
    min_window = min_window

    max_window = data_end - data_start

    if max_window < min_window:
        max_window = min_window

    # It is very time consuming to search every possible window size and every possible starting point.
    # Here we define a subset to search.
    tot_points = max_window
    window_step = min(tot_points // 10, 20)
    data_step = tot_points // 50

    if window_step == 0:
        window_step = 1
    if data_step == 0:
        data_step = 1

    window_list = [
        0
        for k in range(
            int(math.ceil((max_window - min_window) / float(window_step))) + 1
        )
    ]

    for k in range(int(math.ceil((max_window - min_window) / float(window_step)))):
        window_list[k] = min_window + k * window_step

    window_list[-1] = max_window

    num_fits = 0

    for w in window_list:
        num_fits = num_fits + int(
            math.ceil((data_end - w - data_start) / float(data_step))
        )

    if num_fits < 0:
        num_fits = 1

    start_list = [0 for k in range(num_fits)]
    w_list = [0 for k in range(num_fits)]
    q_start_list = [0.0 for k in range(num_fits)]
    q_end_list = [0.0 for k in range(num_fits)]
    rg_list = [0.0 for k in range(num_fits)]
    rger_list = [0.0 for k in range(num_fits)]
    i0_list = [0.0 for k in range(num_fits)]
    i0er_list = [0.0 for k in range(num_fits)]
    qrg_start_list = [0.0 for k in range(num_fits)]
    qrg_end_list = [0.0 for k in range(num_fits)]
    rsqr_list = [0.0 for k in range(num_fits)]
    chi_sqr_list = [0.0 for k in range(num_fits)]
    reduced_chi_sqr_list = [0.0 for k in range(num_fits)]
    corr_coef_list = [0.0 for k in range(num_fits)]

    success = np.zeros(num_fits)

    current_fit = 0
    # This function takes every window size in the window list, stepts it through the data range, and
    # fits it to get the RG and I0. If basic conditions are met, qmin*RG<1 and qmax*RG<1.35, and RG>0.1,
    # We keep the fit.
    for w in window_list:
        for start in range(data_start, data_end - w, data_step):
            x = qs[start : start + w]
            y = il[start : start + w]
            yerr = iler[start : start + w]

            # Remove NaN and Inf values:
            x = x[np.where(np.isfinite(y))]
            yerr = yerr[np.where(np.isfinite(y))]
            y = y[np.where(np.isfinite(y))]

            RG, I0, RGer, I0er, a, b = calcRg(
                x, y, yerr, transform=False, error_weight=error_weight
            )

            if (
                RG > 0.1
                and q[start] * RG < min_qrg
                and q[start + w - 1] * RG < max_qrg
                and RGer / RG <= 1
            ):
                residual = il[start : start + w] - linear_func(
                    qs[start : start + w], a, b
                )

                r_sqr = (
                    1
                    - np.square(residual).sum()
                    / np.square(
                        il[start : start + w] - il[start : start + w].mean()
                    ).sum()
                )

                if r_sqr > 0.15:
                    chi_sqr = np.square((residual) / iler[start : start + w]).sum()

                    # All of my reduced chi_squared values are too small, so I suspect something isn't right with that.
                    # Values less than one tend to indicate either a wrong degree of freedom, or a serious overestimate
                    # of the error bars for the system.
                    dof = w - 2.0
                    reduced_chi_sqr = chi_sqr / dof

                    # Ideally this would be a pvalue, but I'd have to invest in a lot of intrastructure to actually calculate that in a jitted function
                    corr_coef = 1 - spearmanr(residual, qs[start : start + w])

                    start_list[current_fit] = start
                    w_list[current_fit] = w
                    q_start_list[current_fit] = q[start]
                    q_end_list[current_fit] = q[start + w - 1]
                    rg_list[current_fit] = RG
                    rger_list[current_fit] = RGer
                    i0_list[current_fit] = I0
                    i0er_list[current_fit] = I0er
                    qrg_start_list[current_fit] = q[start] * RG
                    qrg_end_list[current_fit] = q[start + w - 1] * RG
                    rsqr_list[current_fit] = r_sqr
                    chi_sqr_list[current_fit] = chi_sqr
                    reduced_chi_sqr_list[current_fit] = reduced_chi_sqr
                    corr_coef_list[current_fit] = corr_coef

                    success[current_fit] = 1

            current_fit = current_fit + 1

    if np.sum(success) > 0:
        fit_array = np.array(
            [
                [
                    start_list[k],
                    w_list[k],
                    q_start_list[k],
                    q_end_list[k],
                    rg_list[k],
                    rger_list[k],
                    i0_list[k],
                    i0er_list[k],
                    qrg_start_list[k],
                    qrg_end_list[k],
                    rsqr_list[k],
                    chi_sqr_list[k],
                    reduced_chi_sqr_list[k],
                    corr_coef_list[k],
                ]
                for k in range(num_fits)
                if success[k] == 1
            ]
        )

        # Now we evaluate the quality of the fits based both on fitting data and on other criteria.

        # Choice of weights is pretty arbitrary, but has been tested against
        # all the data in the SASBDB (as of 11/2020)
        qmaxrg_weight = 1
        qminrg_weight = 1
        rg_frac_err_weight = 1
        i0_frac_err_weight = 1
        r_sqr_weight = 4
        reduced_chi_sqr_weight = 0
        window_size_weight = win_length_weight
        corr_coef_weight = corr_coefht

        weights = np.array(
            [
                qmaxrg_weight,
                qminrg_weight,
                rg_frac_err_weight,
                i0_frac_err_weight,
                r_sqr_weight,
                reduced_chi_sqr_weight,
                window_size_weight,
                corr_coef_weight,
            ]
        )

        quality = np.zeros(len(fit_array))

        max_window_real = float(max(w_list))

        # This iterates through all the fits, and calculates a score. The score is out of 1, 1 being the best, 0 being the worst.
        indices = list(range(len(fit_array)))
        for a in indices:
            k = int(a)  # This is stupid and should not be necessary. Numba bug?

            # Scores all should be 1 based. Reduced chi_square score is not, hence it not being weighted.
            qmaxrg_score = 1 - abs((fit_array[k, 9] - 1.3) / 1.3)
            qminrg_score = 1 - fit_array[k, 8]
            rg_frac_err_score = 1 - fit_array[k, 5] / fit_array[k, 4]
            i0_frac_err_score = 1 - fit_array[k, 7] / fit_array[k, 6]
            r_sqr_score = fit_array[k, 10]
            reduced_chi_sqr_score = 1 / fit_array[k, 12]  # Not right
            window_size_score = fit_array[k, 1] / max_window_real
            corr_coef_score = fit_array[k, 13]

            scores = np.array(
                [
                    qmaxrg_score,
                    qminrg_score,
                    rg_frac_err_score,
                    i0_frac_err_score,
                    r_sqr_score,
                    reduced_chi_sqr_score,
                    window_size_score,
                    corr_coef_score,
                ]
            )

            total_score = (weights * scores).sum() / weights.sum()

            quality[k] = total_score

            # all_scores[k] = scores

        # I have picked an aribtrary threshold here. Not sure if 0.6 is a good quality cutoff or not.
        if quality.max() > quality_thresh:
            if not single_fit:
                idx = quality.argmax()
                rger = fit_array[:, 5][quality > quality[idx] - 0.1].std()
                i0er = fit_array[:, 7][quality > quality[idx] - 0.1].std()
                idx_min = int(fit_array[idx, 0])
                idx_max = int(fit_array[idx, 0] + fit_array[idx, 1] - 1)
            else:
                idx = quality.argmax()
                idx_min = int(fit_array[idx, 0])
                idx_max = int(fit_array[idx, 0] + fit_array[idx, 1] - 1)

            # Now refine the range a bit
            max_quality = quality.max()
            qual = max_quality

            idx_max_ref = idx_max

            if max_qrg == 1.35:
                max_qrg_ref = 1.3
            else:
                max_qrg_ref = max_qrg

            if q[idx_max] * RG < 1.0:
                quality_scale = 0.9
                r_thresh = 0.1
            else:
                quality_scale = 0.97
                r_thresh = 0.15

            # Refine upper end of range
            while qual > quality_scale * max_quality and idx_max_ref < len(q):
                idx_max_ref = idx_max_ref + 1
                x = qs[idx_min : idx_max_ref + 1]
                y = il[idx_min : idx_max_ref + 1]
                yerr = iler[idx_min : idx_max_ref + 1]

                # Remove NaN and Inf values:
                x = x[np.where(np.isfinite(y))]
                yerr = yerr[np.where(np.isfinite(y))]
                y = y[np.where(np.isfinite(y))]

                RG, I0, RGer, I0er, a, b = calcRg(
                    x, y, yerr, transform=False, error_weight=error_weight
                )

                if (
                    RG > 0.1
                    and q[idx_min] * RG < min_qrg
                    and q[idx_max_ref] * RG < max_qrg_ref
                    and RGer / RG <= 1
                ):
                    residual = il[idx_min : idx_max_ref + 1] - linear_func(
                        qs[idx_min : idx_max_ref + 1], a, b
                    )

                    r_sqr = (
                        1
                        - np.square(residual).sum()
                        / np.square(
                            il[idx_min : idx_max_ref + 1]
                            - il[idx_min : idx_max_ref + 1].mean()
                        ).sum()
                    )

                    if r_sqr > r_thresh:
                        chi_sqr = np.square(
                            residual / iler[idx_min : idx_max_ref + 1]
                        ).sum()

                        # All of my reduced chi_squared values are too small, so I suspect something isn't right with that.
                        # Values less than one tend to indicate either a wrong degree of freedom, or a serious overestimate
                        # of the error bars for the system.
                        dof = w - 2.0
                        reduced_chi_sqr = chi_sqr / dof

                        corr_coef = 1 - spearmanr(
                            residual, qs[idx_min : idx_max_ref + 1]
                        )

                        qmaxrg_score = 1 - abs((q[idx_max_ref] * RG - 1.3) / 1.3)
                        qminrg_score = 1 - q[idx_min] * RG
                        rg_frac_err_score = 1 - RGer / RG
                        i0_frac_err_score = 1 - I0er / I0
                        r_sqr_score = r_sqr
                        reduced_chi_sqr_score = 1 / reduced_chi_sqr  # Not right
                        window_size_score = fit_array[k, 1] / max_window_real
                        corr_coef_score = corr_coef

                        scores = np.array(
                            [
                                qmaxrg_score,
                                qminrg_score,
                                rg_frac_err_score,
                                i0_frac_err_score,
                                r_sqr_score,
                                reduced_chi_sqr_score,
                                window_size_score,
                                corr_coef_score,
                            ]
                        )

                        qual = (weights * scores).sum() / weights.sum()

                        if q[idx_max] * RG < 1.0:
                            quality_scale = 0.9
                            r_thresh = 0.1
                        else:
                            quality_scale = 0.97
                            r_thresh = 0.15

                    else:
                        qual = -1

                else:
                    qual = -1

                max_quality = max(max_quality, qual)

            idx_max = idx_max_ref - 1

            # Refine lower end of range
            idx_min_ref = idx_min
            qual = max_quality

            while qual > 0.97 * max_quality and idx_min_ref > 0:
                idx_min_ref = idx_min_ref - 1
                x = qs[idx_min_ref : idx_max + 1]
                y = il[idx_min_ref : idx_max + 1]
                yerr = iler[idx_min_ref : idx_max + 1]

                # Remove NaN and Inf values:
                x = x[np.where(np.isfinite(y))]
                yerr = yerr[np.where(np.isfinite(y))]
                y = y[np.where(np.isfinite(y))]

                RG, I0, RGer, I0er, a, b = calcRg(
                    x, y, yerr, transform=False, error_weight=error_weight
                )

                if (
                    RG > 0.1
                    and q[idx_min_ref] * RG < min_qrg
                    and q[idx_max] * RG < max_qrg_ref
                    and RGer / RG <= 1
                ):
                    residual = il[idx_min_ref : idx_max + 1] - linear_func(
                        qs[idx_min_ref : idx_max + 1], a, b
                    )

                    r_sqr = (
                        1
                        - np.square(residual).sum()
                        / np.square(
                            il[idx_min_ref : idx_max + 1]
                            - il[idx_min_ref : idx_max + 1].mean()
                        ).sum()
                    )

                    if r_sqr > 0.15:
                        chi_sqr = np.square(
                            (residual) / iler[idx_min_ref : idx_max + 1]
                        ).sum()

                        # All of my reduced chi_squared values are too small, so I suspect something isn't right with that.
                        # Values less than one tend to indicate either a wrong degree of freedom, or a serious overestimate
                        # of the error bars for the system.
                        dof = w - 2.0
                        reduced_chi_sqr = chi_sqr / dof

                        corr_coef = 1 - spearmanr(
                            residual, qs[idx_min_ref : idx_max + 1]
                        )

                        qmaxrg_score = 1 - abs((q[idx_max] * RG - 1.3) / 1.3)
                        qminrg_score = 1 - q[idx_min_ref] * RG
                        rg_frac_err_score = 1 - RGer / RG
                        i0_frac_err_score = 1 - I0er / I0
                        r_sqr_score = r_sqr
                        reduced_chi_sqr_score = 1 / reduced_chi_sqr  # Not right
                        window_size_score = fit_array[k, 1] / max_window_real
                        corr_coef_score = corr_coef

                        scores = np.array(
                            [
                                qmaxrg_score,
                                qminrg_score,
                                rg_frac_err_score,
                                i0_frac_err_score,
                                r_sqr_score,
                                reduced_chi_sqr_score,
                                window_size_score,
                                corr_coef_score,
                            ]
                        )

                        qual = (weights * scores).sum() / weights.sum()
                    else:
                        qual = -1

                else:
                    qual = -1

                max_quality = max(max_quality, qual)

            if idx_min_ref == 0 and qual != -1:
                idx_min = idx_min_ref
            else:
                idx_min = idx_min_ref + 1

            # Recalculate Guinier values with the new min and max idx
            x = qs[idx_min : idx_max + 1]
            y = il[idx_min : idx_max + 1]
            yerr = iler[idx_min : idx_max + 1]

            # Remove NaN and Inf values:
            x = x[np.where(np.isfinite(y))]
            yerr = yerr[np.where(np.isfinite(y))]
            y = y[np.where(np.isfinite(y))]

            rg, i0, Rger, I0er, a, b = calcRg(
                x, y, yerr, transform=False, error_weight=error_weight
            )

            if single_fit:
                rger = Rger
                i0er = I0er

        else:
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1

    else:
        rg = -1
        rger = -1
        i0 = -1
        i0er = -1
        idx_min = -1
        idx_max = -1

    idx_min = idx_min + qmin
    idx_max = idx_max + qmin

    # returns Rg, Rg error, I0, I0 error, the index of the first q point of the fit and the index of the last q point of the fit
    return rg, rger, i0, i0er, idx_min, idx_max


@jit(nopython=True, cache=True)
def rankdata(array):
    temp = array.argsort()
    ranks = np.empty_like(temp)
    ranks[temp] = np.arange(len(array))

    return ranks


@jit(nopython=True, cache=True)
def spearmanr(array1, array2):
    rank1 = rankdata(array1)
    rank2 = rankdata(array2)

    n = rank1.size

    dsq = (rank1 - rank2) ** 2

    rho = 1.0 - (6 * dsq.sum()) / (n * (n**2 - 1))

    # t = rho*np.sqrt((n-2)/(1-rho**2)) #Calcuates t value for student t test

    return rho



import os
import re
import numpy as np
from . import sasm as SASM

# import sasm as SASM


# From SourceL RAWAPI
def load_profiles(filename_list, settings=None):
    """
    Loads individual scattering profiles from text files. This could be
    .dat files, but other file types such as .fit, .fir, .int, or .csv
    can also be loaded. This is a convenience wrapper for
    :py:func:`load_files` that only returns profiles. It should not be used
    for images, instead use :py:func:`load_and_integrate_images`.

    Parameters
    ----------
    filename_list: list
        A list of strings containing the full path to each profile to be loaded
        in.
    settings: :class:`bioxtasraw.RAWSettings.RAWSettings`, optional
        The RAW settings to be used when loading in the files,
        such as the calibration values used when radially averaging images.
        Default is none, this is commonly not used.

    Returns
    -------
    profile_list: list
        A list of individual scattering profile (:class:`bioxtasraw.SASM.SASM`)
        items loaded in, including those obtained from radially averaging any
        images.
    """
    if settings is None:
        settings = None  # __default_settings
        pass

    profile_list, iftm_list, secm_list = load_files(
        filename_list, settings
    )  # profile_list, iftm_list, secm_list, img_list = load_files(filename_list, settings)

    return profile_list


def load_files(filename_list, settings, return_all_images=False):
    """
    Loads all types of files that RAW knows how to load. If images are
    included in the list, then the images are radially averaged as part
    of being loaded in.

    Parameters
    ----------
    filename_list: list
        A list of strings containing the full path to each file to be
        loaded in.
    settings: :class:`bioxtasraw.RAWSettings.RAWSettings`
        The RAW settings to be used when loading in the files, such as the
        calibration values used when radially averaging images.
    return_all_images: bool
        If True, all loaded images are returned. If false, only the first loaded
        image of the last file is returned. Useful for minimizing memory use
        if loading and processing a large number of images. False by default.

    Returns
    -------
    profile_list: list
        A list of individual scattering profile (:class:`bioxtasraw.SASM.SASM`)
        items loaded in, including those obtained from radially averaging any
        images.
    ift_list: list
        A list of individual IFT (:class:`bioxtasraw.SASM.IFTM`) items loaded in.
    series_list: list
        A list of individual series (:class:`bioxtasraw.SECM.SECM`) items
        loaded in.
    img_list: list
        A list of individual images (:class:`numpy.array`) loaded in.
    """

    if not isinstance(filename_list, list):
        filename_list = [filename_list]

    profile_list = []
    ift_list = []
    series_list = []
    img_list = []

    for filename in filename_list:
        filename = os.path.abspath(os.path.expanduser(filename))

        file_ext = os.path.splitext(filename)[1]

        is_profile = False

        # if file_ext == '.sec': #Only use .dat files for SAXS_Assistant
        #     secm = SASFileIO.loadSeriesFile(filename, settings)
        #     series_list.append(secm)

        if file_ext == ".ift" or file_ext == ".out":  # changed from elif
            iftm, img = loadFile(filename, settings, return_all_images=False)

            if isinstance(iftm, list):
                ift_list.append(iftm[0])

        # elif file_ext == '.hdf5':
        #     try:
        #         secm = SASFileIO.loadSeriesFile(filename, settings)
        #         series_list.append(secm)
        #     except Exception:
        #         is_profile = True

        else:
            is_profile = True

        if is_profile:
            sasm, img = loadFile(
                filename,
                settings,  # sasm, img = SASFileIO.loadFile(filename, settings, got rid of IO since imported function needed
                return_all_images=return_all_images,
            )

            if img is not None:
                start_point = settings.get("StartPoint")
                end_point = settings.get("EndPoint")

                if not isinstance(sasm, list):
                    qrange = (start_point, len(sasm.getRawQ()) - end_point)
                    sasm.setQrange(qrange)
                else:
                    qrange = (start_point, len(sasm[0].getRawQ()) - end_point)
                    for each_sasm in sasm:
                        each_sasm.setQrange(qrange)

                if isinstance(img, list):
                    if not return_all_images and len(img_list) == 0:
                        img_list.append(img[0])
                    elif not return_all_images:
                        img_list[0] = img[0]
                    else:
                        img_list.extend(img)
                else:
                    if not return_all_images and len(img_list) == 0:
                        img_list.append(img)
                    elif not return_all_images:
                        img_list[0] = img
                    else:
                        img_list.append(img)

            if isinstance(sasm, list):
                profile_list.extend(sasm)
            else:
                profile_list.append(sasm)

    return profile_list, ift_list, series_list  # , img_lis #No images so commented out


# From Source: SASFileIO
def loadFile(filename, raw_settings, no_processing=False, return_all_images=True):
    """Loads a file an returns a SAS Measurement Object (SASM) and the full image if the
    selected file was an Image file

     NB: This is the function used to load any type of file in RAW
    """
    try:
        file_type = checkFileType(filename)
        # print file_type
    except IOError:
        raise
    except Exception as msg:
        print(str(msg))
        file_type = None

    if file_type == "hdf5":
        try:
            hdf5_file = fabio.open(filename)
            file_type = "image"
        except Exception:
            pass
    else:
        hdf5_file = None

    if file_type == "image":  # No images
        try:
            sasm, img = loadImageFile(
                filename, raw_settings, hdf5_file, return_all_images
            )
        except (ValueError, AttributeError) as msg:
            raise SASExceptions.UnrecognizedDataFormat(
                "No data could be retrieved from the file, unknown format."
            )
            traceback.print_exc()
        except Exception:
            traceback.print_exc()
            raise

        # Always do some post processing for image files
        if not isinstance(sasm, list):
            sasm = [sasm]

        for current_sasm in sasm:
            postProcessProfile(current_sasm, raw_settings, no_processing)

    elif file_type == "hdf5":
        sasm = loadHdf5File(filename, raw_settings)
        img = None

    else:
        sasm = loadAsciiFile(filename, file_type)
        img = None

        # If you don't want to post process asci files, return them as a list
        if not isinstance(sasm, list):
            SASM.postProcessSasm(sasm, raw_settings)

    if not isinstance(sasm, list) and (sasm is None or len(sasm.i) == 0):
        raise SASExceptions.UnrecognizedDataFormat(
            "No data could be retrieved from the file, unknown format."
        )

    return sasm, img


# ----ALT from CHat Remove if dont end up using_________________________
# def loadFile(filename, raw_settings, no_processing=False, return_all_images=True):
#     ''' Loads a file and returns a SAS Measurement Object (SASM) and the full image if the
#         selected file was an Image file
#     '''
#     try:
#         file_type = checkFileType(filename)
#     except IOError:
#         raise
#     except Exception as msg:
#         print(str(msg))
#         file_type = None

#     if file_type not in ['hdf5', 'image']:  # We are not using images
#         sasm = loadAsciiFile(filename, file_type)  # Ensure this function exists
#         img = None
#     else:
#         sasm = None
#         img = None

#     if not isinstance(sasm, list) and (sasm is None or len(sasm.i) == 0):
#         raise ValueError('No data could be retrieved from the file, unknown format.')

#     return sasm, img

# _________________________________________________


def loadAsciiFile(filename, file_type):
    ascii_formats = {
        "rad": loadRadFile,
        "new_rad": loadNewRadFile,
        "primus": loadDatFile,
        # 'bift'       : loadBiftFile, #'ift' is used instead
        "int": loadIntFile,
        "abs": loadIntFile,
        "fit": loadFitFile,
        "fir": loadFitFile,
        "ift": loadIftFile,
        "csv": loadTxtFile,
        "out": loadOutFile,
        "txt": loadTxtFile,
    }

    if file_type is None:
        return None

    sasm = None

    if file_type in ascii_formats:
        sasm = ascii_formats[file_type](filename)

    if sasm is not None and file_type != "ift" and file_type != "out":
        if not isinstance(sasm, list) and len(sasm.i) == 0:
            sasm = None

    if file_type == "rad" and sasm is None:
        sasm = ascii_formats["new_rad"](filename)

        if sasm is None:
            sasm = ascii_formats["primus"](filename)

    if file_type == "primus" and sasm is None:
        sasm = ascii_formats["txt"](filename)

    if sasm is not None and not isinstance(sasm, list):
        sasm.setParameter("filename", os.path.split(filename)[1])

    return sasm


def loadDatFile(filename):
    """Loads a .dat format file"""

    with open(filename, "r") as f:
        lines = f.readlines()

    if len(lines) == 0:
        raise SASExceptions.UnrecognizedDataFormat(
            "No data could be retrieved from the file."
        )

    sasm = makeDatFile(lines, filename)

    return sasm


# ---Need to Attempt new version for as SASDAA2.dat failed due to encoding error. but for now will just note
#


def loadRadFile(filename):
    """NOTE : THIS IS THE OLD RAD FORMAT.."""
    """ Loads a .rad file into a SASM object and attaches the filename and header into the parameters  """

    iq_pattern = four_col_fit
    param_pattern = re.compile(r"[a-zA-Z0-9_]*\s*[:]\s+.*")

    i = []
    q = []
    err = []
    parameters = {"filename": os.path.split(filename)[1]}

    fileheader = {}

    with open(filename, "r") as f:
        for line in f:
            iq_match = iq_pattern.match(line)
            param_match = param_pattern.match(line)

            if iq_match:
                found = iq_match.group().split()
                q.append(float(found[0]))

                i.append(float(found[1]))

                err.append(float(found[2]))

            if param_match:
                found = param_match.group().split()

                if len(found) == 3:
                    try:
                        val = float(found[2])
                    except ValueError:
                        val = found[2]

                    fileheader[found[0]] = val

                elif len(found) > 3:
                    arr = []
                    for each in range(2, len(found)):
                        try:
                            val = float(found[each])
                        except ValueError:
                            val = found[each]

                        arr.append(val)

                    fileheader[found[0]] = arr
                else:
                    fileheader[found[0]] = ""

    parameters = {"filename": os.path.split(filename)[1], "fileHeader": fileheader}

    i = np.array(i)
    q = np.array(q)
    err = np.array(err)

    return SASM.SASM(i, q, err, parameters)


def loadNewRadFile(filename):
    """NOTE : This is a load function for the new rad format"""
    """ Loads a .rad file into a SASM object and attaches the filename and header into the parameters  """

    iq_pattern = three_col_fit
    param_pattern = re.compile(r"[a-zA-Z0-9_]*\s*[:]\s+.*")

    i = []
    q = []
    err = []
    parameters = {"filename": os.path.split(filename)[1]}

    fileheader = {}

    with open(filename, "r") as f:
        for line in f:
            iq_match = iq_pattern.match(line)
            param_match = param_pattern.match(line)

            if iq_match:
                found = iq_match.group().split()
                q.append(float(found[0]))

                i.append(float(found[1]))

                err.append(float(found[2]))

            if param_match:
                found = param_match.group().split()

                if len(found) == 3:
                    try:
                        val = float(found[2])
                    except ValueError:
                        val = found[2]

                    fileheader[found[0]] = val

                elif len(found) > 3:
                    arr = []
                    for each in range(2, len(found)):
                        try:
                            val = float(found[each])
                        except ValueError:
                            val = found[each]

                        arr.append(val)

                    fileheader[found[0]] = arr
                else:
                    fileheader[found[0]] = ""

    parameters = {"filename": os.path.split(filename)[1], "counters": fileheader}

    i = np.array(i)
    q = np.array(q)
    err = np.array(err)

    return SASM.SASM(i, q, err, parameters)


def loadIntFile(filename):
    """Loads a simulated SAXS data curve .int or .abs file"""
    fit_list = [two_col_fit, three_col_fit, four_col_fit, five_col_fit, seven_col_fit]

    i = []
    q = []
    err = []

    with open(filename, "r") as f:
        firstLine = f.readline()

        match = [fit.match(firstLine) for fit in fit_list]

        if any(match):
            fileHeader = {}
        else:
            fileHeader = _parse_header_line(firstLine)

        parameters = {"filename": os.path.split(filename)[1], "counters": fileHeader}

        if len(fileHeader) == 0:
            q, i = _match_int_lines(firstLine, q, i, fit_list)

        for line in f:
            q, i = _match_int_lines(line, q, i, fit_list)

    i = np.array(i)
    q = np.array(q)
    err = np.sqrt(abs(i))

    return SASM.SASM(i, q, err, parameters)


def loadTxtFile(filename):
    """Loads a generic two or three column text file with space, tab, or comma separated values"""
    fit_list = [three_col_fit, i_q_err_match, two_col_fit, i_q_match]

    i = []
    q = []
    err = []

    with open(filename, "r") as f:
        firstLine = f.readline()

        match = [fit.match(firstLine) for fit in fit_list]

        if any(match):
            fileHeader = {}
        else:
            fileHeader = {"comment": firstLine}
            firstline_l = firstLine.lower()
            if "chi^2" in firstline_l:
                chisq = firstline_l.split("chi^2")[-1].strip(":= ").split()[0].strip()
                fileHeader["Chi_squared"] = float(chisq)

            if "rg" in firstline_l:
                rg = firstline_l.split("rg")[-1].strip("t:= ").split()[0].strip()
                fileHeader["Rg"] = float(rg)

            if "dro" in firstline_l:
                dro = firstline_l.split("dro")[-1].strip(":= ").split()[0].strip()
                fileHeader["Hydration_shell_contrast"] = float(dro)

            if "vol" in firstline_l:
                vol = firstline_l.split("vol")[-1].strip(":= ").split()[0].strip()
                fileHeader["Excluded_volume"] = float(vol)

        parameters = {"filename": os.path.split(filename)[1], "counters": fileHeader}

        if len(fileHeader) == 0:
            q, i, err = _match_txt_lines(firstLine, q, i, err, fit_list)

        for line in f:
            q, i, err = _match_txt_lines(line, q, i, err, fit_list)

    i = np.array(i)
    q = np.array(q)

    if len(err) == len(i):
        err = np.array(err)
    else:
        err = np.sqrt(abs(i))

    return SASM.SASM(i, q, err, parameters)


def loadFitFile(filename):
    i = []
    q = []
    err = []
    fit = []

    with open(filename, "r") as f:
        firstLine = f.readline()

        three_col_match = three_col_fit.match(firstLine)
        four_col_match = four_col_fit.match(firstLine)
        five_col_match = five_col_fit.match(firstLine)
        if three_col_match or four_col_match or five_col_match:
            fileHeader = {}

        else:
            fileHeader = _parse_header_line(firstLine)

        if "Experimental" in firstLine:
            sasref = True  # SASREFMX Fit file (Damn those hamburg boys and their 50 different formats!)
        else:
            sasref = False

        parameters = {"filename": os.path.split(filename)[1], "counters": fileHeader}

        path_noext, ext = os.path.splitext(filename)

        fit_parameters = {
            "filename": os.path.split(path_noext)[1] + "_FIT",
            "counters": fileHeader,
        }

        if len(fileHeader) == 0:
            q, i, err, fit = _match_fit_lines(firstLine, q, i, err, fit, sasref)

        for line in f:
            q, i, err, fit = _match_fit_lines(line, q, i, err, fit, sasref)

    if len(i) == 0:
        raise SASExceptions.UnrecognizedDataFormat(
            "No data could be retrieved from the file, unknown format."
        )

    q = np.array(q)
    i = np.array(i)
    err = np.array(err)
    fit = np.array(fit)

    fit_sasm = SASM.SASM(fit, np.copy(q), np.copy(err), fit_parameters)

    sasm = SASM.SASM(i, q, err, parameters)

    return [sasm, fit_sasm]


def loadIftFile(filename):
    # Loads RAW BIFT .ift files into IFTM objects
    iq_pattern = four_col_fit
    pr_pattern = three_col_fit
    extrap_pattern = two_col_fit

    r = []
    p = []
    err = []

    q = []
    i = []
    err_orig = []
    fit = []

    q_extrap = []
    fit_extrap = []

    with open(filename, "r") as f:
        path_noext, ext = os.path.splitext(filename)

        for line in f:
            pr_match = pr_pattern.match(line)
            iq_match = iq_pattern.match(line)
            extrap_match = extrap_pattern.match(line)

            if pr_match:
                found = pr_match.group().split()

                r.append(float(found[0]))
                p.append(float(found[1]))
                err.append(float(found[2]))

            elif iq_match:
                found = iq_match.group().split()

                q.append(float(found[0]))
                i.append(float(found[1]))
                err_orig.append(float(found[2]))
                fit.append(float(found[3]))

            elif extrap_match:
                found = extrap_match.group().split()

                q_extrap.append(float(found[0]))
                fit_extrap.append(float(found[1]))

    p = np.array(p)
    r = np.array(r)
    err = np.array(err)
    i = np.array(i)
    q = np.array(q)
    err_orig = np.array(err_orig)
    fit = np.array(fit)
    q_extrap = np.array(q_extrap)
    fit_extrap = np.array(fit_extrap)

    # Check to see if there is any header from RAW, and if so get that.
    with open(filename, "r") as f:
        all_lines = f.readlines()

    header = []
    for j in range(len(all_lines)):
        if "### HEADER:" in all_lines[j]:
            header = all_lines[j + 1 :]

    hdict = {}

    if len(header) > 0:
        hdr_str = ""
        for each_line in header:
            hdr_str = hdr_str + each_line.lstrip("#")
        try:
            hdict = dict(json.loads(hdr_str))
        except Exception:
            pass

    parameters = hdict
    parameters["filename"] = os.path.split(filename)[1]

    if q.size == 0:
        q = np.array([0, 0])
        i = q
        err_orig = q
        fit = q

    if q_extrap.size == 0:
        q_extrap = q
        fit_extrap = fit

    iftm = SASM.IFTM(p, r, err, i, q, err_orig, fit, parameters, fit_extrap, q_extrap)

    return [iftm]


def loadOutFile(filename):
    with open(filename, "r") as f:
        lines = f.readlines()

    iftm = parse_out_file(lines)

    iftm.setParameter("filename", os.path.basename(filename))

    return [iftm]


def checkFileType(filename):
    """Tries to find out what file type it is and reports it back"""

    path, ext = os.path.splitext(filename)

    if ext == ".fit":
        return "fit"
    elif ext == ".fir":
        return "fir"
    elif ext == ".abs":
        return "abs"
    elif ext == ".out":
        return "out"
    elif ext == ".nxs":  # Nexus file
        return "image"
    elif ext == ".edf":
        return "image"
    elif ext == ".ccdraw":
        return "image"
    elif ext == ".int":
        return "int"
    elif (
        ext == ".img"
        or ext == ".imx_0"
        or ext == ".dkx_0"
        or ext == ".dkx_1"
        or ext == ".png"
        or ext == ".mpa"
    ):
        return "image"
    elif ext == ".dat" or ext == ".sub" or ext == ".txt":
        return "primus"
    elif (
        ext == ".mar1200" or ext == ".mar2400" or ext == ".mar2300" or ext == ".mar3600"
    ):
        return "image"
    elif (
        ext == ".img"
        or ext == ".sfrm"
        or ext == ".dm3"
        or ext == ".edf"
        or ext == ".xml"
        or ext == ".cbf"
        or ext == ".kccd"
        or ext == ".msk"
        or ext == ".spr"
        or ext == ".tif"
        or ext == ".mccd"
        or ext == ".mar3450"
        or ext == ".npy"
        or ext == ".pnm"
        or ext == ".No"
    ):
        return "image"
    elif ext == ".ift":
        return "ift"
    elif ext == ".csv":
        return "txt"
    elif ext == ".h5":
        return "hdf5"
    elif ext == ".pdb":
        return "pdb"
    elif ext == ".cif":
        return "cif"
    else:
        try:
            f = fabio.open(filename)
            return "image"
        except Exception:
            try:
                float(ext.strip("."))
            except Exception:
                return "txt"
            return "csv"


def makeDatFile(lines, filename):
    iq_pattern = i_q_err_match

    i = []
    q = []
    err = []

    comment = ""
    line = lines[0]
    j = 0
    while line.split() and line.strip()[0] == "#":
        comment = comment + line
        j = j + 1
        line = lines[j]

    fileHeader = {"comment": comment}
    parameters = {"filename": os.path.split(filename)[1], "counters": fileHeader}

    if comment.find("model_intensity") > -1:
        # FoXS file with a fit! has four data columns
        is_foxs_fit = True
        is_sans_data = False
        imodel = []

    elif comment.find("dQ") > -1:
        # ORNL SANS instrument file
        is_foxs_fit = False
        is_sans_data = True
        qerr = []
    else:
        is_foxs_fit = False
        is_sans_data = False

    header = []
    header_start = False

    for j, line in enumerate(lines):
        iq_match = iq_pattern.match(line)

        if iq_match:
            if is_foxs_fit:
                if "," in line:
                    found = line.split(",")
                else:
                    found = line.split()
                q.append(float(found[0]))
                i.append(float(found[1]))
                imodel.append(float(found[2]))
                err.append(abs(float(found[3])))

            elif is_sans_data:
                found = iq_match.group()
                if "," in found:
                    found = line.split(",")
                else:
                    found = line.split()
                q.append(float(found[0]))
                i.append(float(found[1]))
                err.append(abs(float(found[2])))
                qerr.append(abs(float(found[3])))

            else:
                found = iq_match.group()
                if "," in found:
                    found = found.split(",")
                else:
                    found = found.split()
                q.append(float(found[0]))
                i.append(float(found[1]))
                err.append(abs(float(found[2])))

        # Check to see if there is any header from RAW, and if so get that.
        # Header at the bottom
        if "### HEADER:" in line and len(q) > 0:
            header = lines[j + 1 :]

            # For headers at the bottom, stop trying the regex
            if len(q) > 0:
                break

        # Header at top
        elif "### HEADER:" in line and len(q) == 0:
            header_start = True

        elif header_start and "### DATA:" in line:
            header_start = False

        elif header_start and not iq_match:
            header.append(lines[j])

    if len(header) > 0:
        hdr_str = ""
        for each_line in header:
            hdr_str = hdr_str + each_line.lstrip("#")

        hdict = loadDatHeader(hdr_str)

        for each in hdict:
            if each != "filename":
                parameters[each] = hdict[each]

    i = np.array(i)
    q = np.array(q)
    err = np.array(err)

    sasm = SASM.SASM(i, q, err, parameters)

    if is_foxs_fit:
        parameters2 = copy.copy(parameters)
        parameters2["filename"] = (
            os.path.splitext(os.path.split(filename)[1])[0] + "_FIT"
        )

        sasm_model = SASM.SASM(imodel, q, err, parameters2)

        return [sasm, sasm_model]

    elif is_sans_data:
        sasm.setRawQErr(np.array(qerr))
        sasm._update()

    return sasm


def postProcessProfile(sasm, raw_settings, no_processing):
    """
    Does post-processing on profiles created from images.
    """
    SASM.postProcessSasm(sasm, raw_settings)  # Does dezingering

    if not no_processing:
        # Need to do a little work before we can do glassy carbon normalization
        if raw_settings.get("NormAbsCarbon") and not raw_settings.get(
            "NormAbsCarbonIgnoreBkg"
        ):
            bkg_filename = raw_settings.get("NormAbsCarbonSamEmptyFile")
            bkg_sasm = raw_settings.get("NormAbsCarbonSamEmptySASM")
            if (
                bkg_sasm is None
                or bkg_sasm.getParameter("filename") != os.path.split(bkg_filename)[1]
            ):
                bkg_sasm, junk_img = loadFile(
                    bkg_filename, raw_settings, no_processing=True
                )
                if isinstance(bkg_sasm, list):
                    if len(bkg_sasm) > 1:
                        bkg_sasm = SASProc.average(bkg_sasm, copy_params=False)
                    else:
                        bkg_sasm = bkg_sasm[0]
                raw_settings.set("NormAbsCarbonSamEmptySASM", bkg_sasm)

        try:
            # Does fully glassy carbon abs scale
            SASCalib.postProcessImageSasm(sasm, raw_settings)
        except SASExceptions.AbsScaleNormFailed:
            raise


# Define regular expressions for ascii files
start_match = r"\s*"
end_match = r"\s*$"
num_match = r"[\+-]?\d+\.?\d*[+eE-]*\d*"
sep_match = r"\s*[\s,]\s*"
alt_sep_match = r"\s*[\s,]?\s*"

two_col_fit = re.compile(
    start_match + (num_match + sep_match) * 1 + num_match + end_match
)
i_q_match = re.compile(
    start_match + (num_match + sep_match) * 1 + num_match + alt_sep_match
)
three_col_fit = re.compile(
    start_match + (num_match + sep_match) * 2 + num_match + end_match
)
i_q_err_match = re.compile(
    start_match + (num_match + sep_match) * 2 + num_match + alt_sep_match
)
four_col_fit = re.compile(
    start_match + (num_match + sep_match) * 3 + num_match + end_match
)
five_col_fit = re.compile(
    start_match + (num_match + sep_match) * 4 + num_match + end_match
)
seven_col_fit = re.compile(
    start_match + (num_match + sep_match) * 6 + num_match + end_match
)


# ----- Added for SASDD22
def loadDatHeader(header):
    try:
        hdict = dict(json.loads(header))
        # print 'Loading RAW info/analysis...'
    except Exception:
        # print 'Unable to load header/analysis information. Maybe the file was not generated by RAW or was generated by an old version of RAW?'
        hdict = {}

    if hdict:
        hdict = translateHeader(hdict, to_sasbdb=False)

    return hdict


def translateHeader(header, to_sasbdb=True):
    """
    Translates the header keywords to or from matching SASBDB format. This is
    to add compatibility with SASBDB while maintaining compatibility with older
    RAW formats and RAW internals.
    """
    new_header = copy.deepcopy(header)

    for key in header.keys():
        if isinstance(header[key], dict):
            new_header[key] = translateHeader(header[key], to_sasbdb)
        else:
            if to_sasbdb:
                if key in sasbdb_trans:
                    new_header[sasbdb_trans[key]] = new_header.pop(key)
            else:
                if key in sasbdb_back_trans:
                    new_header[sasbdb_back_trans[key]] = new_header.pop(key)

    return new_header


sasbdb_trans = {
    # First general RAW keywords
    "Sample_Detector_Distance": "Sample-to-detector distance (mm)",
    "Wavelength": "Wavelength (A)",
    # Next BioCAT specific keywords
    "Exposure_time/frame_s": "Exposure time/frame (s)",
    "LC_flow_rate_mL/min": "Flow rate (ml/min)",
}

sasbdb_back_trans = {value: key for (key, value) in sasbdb_trans.items()}


def _match_txt_lines(line, q, i, err, fit_list):
    for fit in fit_list:
        match = fit.match(line)

        if match:
            data = match.group()
            if "," in data:
                data = data.split(",")
            else:
                data = data.split()

            q.append(float(data[0]))
            i.append(float(data[1]))

            if len(data) == 3:
                try:
                    err.append(float(data[2]))
                except Exception:
                    pass

            break

    return q, i, err



"""
This module provides tools for working with PDDF (Pair Distribution Function) data and selection.
This includes functions to characterize the data like the Elongation ratio from Putnam 2016
"""

import pandas as pd
from scipy.integrate import trapezoid
import numpy as np
import warnings
import logging
from .rawutils import rawapitry as raw

# ******************************************************************************
# This file is part of SAXS Assistant.
#
#    All code from SAXS Assisstant can be used freely for non-commercial purposes.
#    If you use this code in your work, please cite the following publications:
# Add The GPA citation
# Add Franke citation
# Add Raw citation
# SAXS Assistant citation
#    SAXS Assistant is based on the code from RAW, which is a SAXS data analysis software.
#
#
#    SAXS Assistant utilizes parts of the code from RAW
#    SAXS Assistant is shared for helping the community with SAXS data analysis.
#    but it is not a replacement for RAW, and does not include all the features of RAW.
#    SAXS Assisant does not offer warranty-- use at your own risk and evaluate the results carefully.


#    RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with RAW.  If not, see <http://www.gnu.org/licenses/>.
#
# ******************************************************************************
def get_ER(p, r):
    """
    Calculate the Elongation Ratio (ER) from a PDDF.
    If available, some returned values from BIFT cause issues with the
    calculation of the ER, so if this is the case, this isn't calculated.
    Parameters
      ----------
      p : numpy.ndarray
        The PDDF values, P(r)
      r : numpy.ndarray
        The distance values, r
    """
    pr_peak = pd.Series(p, name="P(r)").idxmax()
    P_r = pd.DataFrame({"r": r, "P(r)": p})

    r_spacing = []
    last_r = 0
    current_r = 0
    for j in range(len(P_r["r"])):
        current_r = P_r["r"][j]
        r_spacing.append(current_r - last_r)

        last_r = current_r
    r_spacing = pd.Series(r_spacing, name="delta r")
    P_r = pd.concat([P_r, r_spacing], axis=1)

    zero_R = trapezoid(
        P_r["P(r)"][: pr_peak + 1],
        x=P_r["r"][: pr_peak + 1],
        dx=P_r["delta r"][1],
        axis=-1,
    )
    R_dmax = trapezoid(
        P_r["P(r)"][pr_peak:], x=P_r["r"][pr_peak:], dx=P_r["delta r"][1], axis=-1
    )
    ER = R_dmax / zero_R
    return ER


def unpack_pr_fits_dict(pr_fits_dict):
    """
    Unpacks a pr_fits_dict into the separate lists needed for ranking.
    Returns:
        rg_min, i0_p_list, chi_sq_list, dmax_mins, logas, dmax_err, pr_list,
        pr_i_orig, pr_fit, pr_err_orig, pr_q_orig, pr_qxt, nmins
    """
    rg_min, i0_p_list, chi_sq_list, dmax_mins = [], [], [], []
    logas, dmax_err = [], []
    pr_rg_err, pr_i0_err, pr_qmin, pr_qmax = [], [], [], []
    pr_list, pr_i_orig, pr_fit, pr_err_orig, pr_q_orig, pr_qxt, nmins = (
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    )

    for nmin, fit in pr_fits_dict.items():
        try:
            rg_min.append(fit["Rg"])
            pr_rg_err.append(fit.get("Rg Err", np.nan))  # New
            i0_p_list.append(fit["I0"])
            pr_i0_err.append(fit.get("I0 Err", np.nan))  # New
            chi_sq_list.append(fit["Chi^2"])
            dmax_mins.append(fit["Dmax"])
            logas.append(fit.get("Log Alpha", 0))
            dmax_err.append(fit.get("Dmax Error", 0))
            pr_qmin.append(fit.get("qmin", np.nan))  # New
            pr_qmax.append(fit.get("qmax", np.nan))  # New
            pr_list.append(fit["p(r)"])
            pr_i_orig.append(fit["i_orig"])
            pr_fit.append(fit["i_fit"])
            pr_err_orig.append(fit["err_orig"])
            pr_q_orig.append(fit["q_orig"])
            pr_qxt.append(fit.get("q_extrap", fit["q_orig"]))
            nmins.append(nmin)
        except KeyError as e:
            logging.warning(f"Missing key in pr_fits_dict[{nmin}]: {e}")
            continue

    return (
        rg_min,
        pr_rg_err,
        i0_p_list,
        pr_i0_err,
        chi_sq_list,
        dmax_mins,
        logas,
        dmax_err,
        pr_qmin,
        pr_qmax,
        pr_list,
        pr_i_orig,
        pr_fit,
        pr_err_orig,
        pr_q_orig,
        pr_qxt,
        nmins,
    )


def get_all_pr_results(profile, q, I, err):
    """
    Runs raw.bift at different nmin values and stores the full results for later evaluation.
    #Here maybe dont need to send q, I, err
    """
    pr_fits = {}
    for nmin in range(0, 26, 5):
        # print(f"Running P(r) at nmin = {nmin}"+ "q = "+str(q[nmin]))
        try:
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            (
                gi_bift,
                gi_bift_dmax,
                gi_bift_rg,
                gi_bift_i0,
                gi_bift_dmax_err,
                gi_bift_rg_err,
                gi_bift_i0_err,
                gi_bift_chi_sq,
                gi_bift_log_alpha,
                gi_bift_log_alpha_err,
                gi_bift_evidence,
                gi_bift_evidence_err,
            ) = raw.bift(profile, idx_min=nmin, pr_pts=100, use_guinier_start=False)

            try:
                er = get_ER(gi_bift.p, gi_bift.r)
            except:
                er = None

            pr_fits[nmin] = {
                "Rg": gi_bift_rg,
                "Rg Err": gi_bift_rg_err,
                "I0 Err": gi_bift_i0_err,
                "I0": gi_bift_i0,
                "Dmax": gi_bift_dmax,
                "Chi^2": gi_bift_chi_sq,
                "Log Alpha": gi_bift_log_alpha,
                "Dmax Error": gi_bift_dmax_err,
                "ER": er,
                "q_orig": gi_bift.q_orig,
                "q_extrap": gi_bift.q_extrap,
                "qmin": np.min(gi_bift.q_orig),
                "qmax": np.max(gi_bift.q_orig),
                "i_fit": gi_bift.i_fit,
                "i_orig": gi_bift.i_orig,
                "err_orig": gi_bift.err_orig,
                "p(r)": (gi_bift.r, gi_bift.p),
            }
        except AttributeError:
            logging.warning(
                f"P(r) AttributeError at nmin = {nmin} for {df_wrong['file name'][j]}"
            )  # Gotta fix this
        except Exception as e:
            logging.warning(
                f"P(r) failed at nmin = {nmin} for {df_wrong['file name'][j]}: {e}"
            )
    return pr_fits


def select_best_pr_fit(
    pr_rg_list,
    pr_rg_err_list,
    pr_i0_list,
    pr_i0_err_list,
    chi_sq_list,
    dmax_list,
    log_alpha_list,
    dmax_err_list,
    pr_qmin_list,
    pr_qmax_list,  # ✅ NEW
    final_guinier_rg,
    final_guinier_i0,
    pr_list,
    pr_i_orig,
    pr_fit,
    pr_err_orig,
    pr_q_orig,
    pr_qxt,
    nmins,
    q,
    I,
    err,
    sample_id,
    murthy_df,
    j,
):
    """
    V2. Selects the best P(r) curve whose Rg is within 15% of the final Guinier Rg.
    Updates murthy_df and plot_data with selected P(r). Prioritizes the Rg similarity, followed by Chi distance to 1,
    then the Dmax error, and then the Log Alpha, the candidate Pr at the top of dataframe is returned.
    """
    import numpy as np

    def abs_percent_error(a, b):
        return abs(a - b) / b * 100

    try:
        # Identify indices of candidate P(r) fits within 15% of Guinier Rg
        rg_diffs = [abs_percent_error(final_guinier_rg, rg) for rg in pr_rg_list]
        candidate_inds = np.argwhere(np.array(rg_diffs) < 15).flatten()

        if len(candidate_inds) == 0:
            murthy_df.loc[murthy_df.index[j], "Flag"] = "No PR match to Rg"
            logging.warning(f"No valid P(r) match for {sample_id}")
            return None

        # Build candidate dataframe
        rows = []
        for i in candidate_inds:
            rows.append(
                [
                    i,
                    dmax_list[i],
                    round(chi_sq_list[i], 3),
                    round(log_alpha_list[i], 2),
                    round(dmax_err_list[i], 3),
                    pr_i0_list[i],
                    final_guinier_i0,
                    pr_rg_list[i],
                    final_guinier_rg,
                    abs(
                        1 - round(chi_sq_list[i], 3),
                    ),
                    abs(pr_rg_list[i] - final_guinier_rg),
                ]
            )
        fin_pr = pd.DataFrame(
            rows,
            columns=[
                "Index",
                "Dmax",
                "Chi^2",
                "Log alpha",
                "Dmax Error",
                "Pr i0",
                "G i0",
                "Pr Rg",
                "G Rg",
                "chi dist",
                "Rg Abs dif",
            ],
        )

        # Sort to select best
        # fin_pr = fin_pr.sort_values(by=['chi dist', 'Dmax Error', 'Log alpha'], ascending=[True, True, False]).reset_index(drop=True)
        fin_pr = fin_pr.sort_values(
            by=["Rg Abs dif", "chi dist", "Dmax Error", "Log alpha"],
            ascending=[True, True, True, False],
        ).reset_index(drop=True)

        # display(fin_pr)
        best_idx = int(fin_pr["Index"][0])

        # Update dataframe
        murthy_df.loc[murthy_df.index[j], "Dmax"] = round(dmax_list[best_idx], 4)
        murthy_df.loc[murthy_df.index[j], "Chi^2"] = round(chi_sq_list[best_idx], 4)
        murthy_df.loc[murthy_df.index[j], "Dmax Error RAW"] = round(
            fin_pr["Dmax Error"][0], 4
        )
        murthy_df.loc[murthy_df.index[j], "Pr Rg"] = round(pr_rg_list[best_idx], 4)
        murthy_df.loc[murthy_df.index[j], "Pr i0"] = pr_i0_list[best_idx]
        murthy_df.loc[murthy_df.index[j], "Pr Log Alpha"] = log_alpha_list[best_idx]
        murthy_df.loc[murthy_df.index[j], "Pr nmin"] = round(nmins[best_idx], 4)
        murthy_df.loc[murthy_df.index[j], "Pr Rg Err"] = round(
            pr_rg_err_list[best_idx], 4
        )
        murthy_df.loc[murthy_df.index[j], "Pr i0 Err"] = round(
            pr_i0_err_list[best_idx], 4
        )
        murthy_df.loc[murthy_df.index[j], "Pr qmin"] = round(pr_qmin_list[best_idx], 4)
        murthy_df.loc[murthy_df.index[j], "Pr qmax"] = round(pr_qmax_list[best_idx], 4)
        return fin_pr, best_idx

    except Exception as e:
        logging.warning(f"Error selecting final P(r) for {sample_id}: {e}")
        murthy_df.loc[murthy_df.index[j], "Flag"] = "PR selection error"
        return None



# -*- coding: utf-8 -*-
"""BIFT

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1Ex2b0GScqziQGV5uwVlRwYgng5ebsiGY
"""

# ******************************************************************************
# This file is part of BioXTAS RAW.
#
#    BioXTAS RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    BioXTAS RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with BioXTAS RAW.  If not, see <http://www.gnu.org/licenses/>.
#
# ******************************************************************************

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import object, range, map, zip

# import os
# from io import open
import six

import multiprocessing
import functools
import threading
import platform

import scipy.optimize
import scipy.interpolate
import numpy as np
from numba import jit

# raw_path = os.path.abspath(os.path.join('.', __file__, '..', '..'))
# if raw_path not in os.sys.path:
#     os.sys.path.append(raw_path)

from . import sasm as SASM  # import bioxtasraw.SASM as SASM


@jit(nopython=True, cache=True)
def createTransMatrix(q, r):
    """
    Matrix such that when you take T dot P(r) you get I(q),
    The A_ij matrix in equation (2) of Hansen 2000.
    """
    T = np.outer(q, r)
    T = 4 * np.pi * (r[1] - r[0]) * np.where(T == 0, 1, np.sin(T) / T)

    return T


@jit(nopython=True, cache=True)
def distDistribution_Sphere(i0_meas, N, dmax):
    """Creates the initial P(r) function for the prior as a sphere.

    Formula from Svergun & Koch 2003:

    p(r) = 4*pi*D**3/24*r**2*(1-1.5*r/D+0.5*(r/D)**3)

    Normalized so that the area is equal to the measured I(0). Integrate p(r):

    I(0) = integral of p(r) from 0 to Dmax
         = (4*pi*D^3/24)**2

    So the normalized p(r) is:

    p(r) = r**2*(1-1.5*r/D+0.5*(r/D)**3) * I(0)_meas/(4*pi*D**3/24)

    To match convention with old RAW BIFT, we also carry around an extra factor
    of 4*pi*Delta r, so the normalization becomes:

    p(r) = r**2*(1-1.5*r/D+0.5*(r/D)**3) * I(0)_meas/(D**3/(24*Delta_r))
    """

    r = np.linspace(0, dmax, N + 1)

    p = r**2 * (1 - 1.5 * (r / dmax) + 0.5 * (r / dmax) ** 3)
    p = p * i0_meas / (4 * np.pi * dmax**3 / 24.0)
    # p = p * i0_meas/(dmax**3/(24.*(r[1]-r[0])))   #Which normalization should I use? I'm not sure either agrees with what Hansen does.

    return p, r


@jit(nopython=True, cache=True)
def makePriorDistribution(i0, N, dmax, dist_type="sphere"):
    if dist_type == "sphere":
        p, r = distDistribution_Sphere(i0, N, dmax)
    else:
        r = np.linspace(0, dmax, N + 1)
        p = np.zeros_like(r)

    return p, r


@jit(nopython=True, cache=True)
def bift_inner_loop(f, p, B, alpha, N, sum_dia):
    # Define starting conditions and loop variables
    ite = 0
    maxit = 2000
    minit = 100
    xprec = 0.999
    dotsp = 0
    omega = 0.5

    sigma = np.zeros_like(p)

    # Start loop
    while ite < maxit and not (ite > minit and dotsp > xprec):
        ite = ite + 1

        # Apply positivity constraint
        sigma[1:-1] = np.abs(p[1:-1] + 1e-10)
        p_neg_idx = p[1:-1] <= 0
        f_neg_idx = f[1:-1] <= 0
        p[1:-1][p_neg_idx] = p[1:-1][p_neg_idx] * -1 + 1e-10
        f[1:-1][f_neg_idx] = f[1:-1][f_neg_idx] * -1 + 1e-10

        # Apply smoothness constraint
        for k in range(2, N - 1):
            p[k] = (f[k - 1] + f[k + 1]) / 2.0

        p[1] = f[2] / 2.0
        p[-2] = p[-3] / 2.0

        p[0] = f[0]
        p[-1] = f[-1]

        sigma[0] = 10

        # Calculate the next correction
        for k in range(1, N):
            fsumi = 0

            for j in range(1, N):
                fsumi = fsumi + B[k, j] * f[j]

            fsumi = fsumi - B[k, k] * f[k]

            fx = (2 * alpha * p[k] / sigma[k] + sum_dia[k] - fsumi) / (
                2 * alpha / sigma[k] + B[k, k]
            )

            f[k] = (1 - omega) * f[k] + omega * fx

        # Calculate convergence
        gradsi = -2 * (f[1:-1] - p[1:-1]) / sigma[1:-1]
        gradci = 2 * (np.sum(B[1:-1, 1:-1] * f[1:-1], axis=1) - sum_dia[1:-1])

        wgrads = np.sqrt(np.abs(np.sum(gradsi**2)))
        wgradc = np.sqrt(np.abs(np.sum(gradci**2)))

        if wgrads * wgradc == 0:
            dotsp = 1
        else:
            dotsp = np.sum(gradsi * gradci) / (wgrads * wgradc)

    return f, p, sigma, dotsp, xprec


@jit(nopython=True, cache=True)
def getEvidence(params, q, i, err, N):
    alpha, dmax = params
    alpha = np.exp(alpha)

    err = err**2

    p, r = makePriorDistribution(
        i[0], N, dmax, "sphere"
    )  # Note, here I use p for what Hansen calls m
    T = createTransMatrix(q, r)

    p[0] = 0
    f = np.zeros_like(p)

    # norm_T = T/err[:,None]  #Slightly faster to create this first
    norm_T = T / err.reshape((err.size, 1))  # Slightly faster to create this first

    # sum_dia = np.sum(norm_T*i[:,None], axis=0)   #Creates YSUM in BayesApp code, some kind of calculation intermediate
    sum_dia = np.sum(
        norm_T * i.reshape((i.size, 1)), axis=0
    )  # Creates YSUM in BayesApp code, some kind of calculation intermediate
    sum_dia[0] = 0

    B = np.dot(T.T, norm_T)  # Creates B(i, j) in BayesApp code
    B[0, :] = 0
    B[:, 0] = 0

    # Do some kind of rescaling of the input
    c1 = np.sum(np.sum(T[1:4, 1:-1] * p[1:-1], axis=1) / err[1:4])
    c2 = np.sum(i[1:4] / err[1:4])
    p[1:-1] = p[1:-1] * (c2 / c1)
    f[1:-1] = p[1:-1] * 1.001  # Note: f is called P in the original RAW BIFT code

    # Do the optimization
    f, p, sigma, dotsp, xprec = bift_inner_loop(f, p, B, alpha, N, sum_dia)

    # Calculate the evidence
    s = np.sum(-((f[1:-1] - p[1:-1]) ** 2) / sigma[1:-1])
    # print('i from bift ', i)
    # print(' ')
    # print('T from bift ', T)
    # print(' ')
    # print('f from bift ---------------------',f)
    # print(' err from bift ', err)
    # print(' ')
    # print('i.size, p.size from bift ', i.size, p.size)
    c = np.sum((i[1:-1] - np.sum(T[1:-1, 1:-1] * f[1:-1], axis=1)) ** 2 / err[1:-1]) / (
        i.size - p.size
    )

    u = np.sqrt(np.abs(np.outer(f[1:-1], f[1:-1]))) * B[1:-1, 1:-1] / alpha

    # u[np.diag_indices(u.shape[0])] = u[np.diag_indices(u.shape[0])]+1
    for j in range(0, u.shape[0]):
        u[j, j] = u[j, j] + 1

    # Absolute value of determinant is equal to the product of the singular values
    rlogdet = np.log(np.abs(np.linalg.det(u)))

    evidence = (
        -np.log(abs(dmax))
        + (alpha * s - 0.5 * c * i.size)
        - 0.5 * rlogdet
        - np.log(abs(alpha))
    )

    # Some kind of after the fact adjustment

    if evidence <= 0 and dotsp < xprec:
        evidence = evidence * 30
    elif dotsp < xprec:
        evidence = evidence / 30.0
    # print('c from BIFT ', c)#--------------------------------------------------------------------------
    return evidence, c, f, r


def getEvidenceOptimize(params, q, i, err, N):
    evidence, c, f, r = getEvidence(params, q, i, err, N)
    # print('c from non jit ',c)
    # print(' f from non jit ', f)
    # Negative so you can minimize on it
    return -evidence


def calc_bift_errors(
    opt_params,
    q,
    i,
    err,
    N,
    mc_runs=300,
    abort_check=False,
    single_proc=False,
    nprocs=0,
):
    # First, randomly generate a set of parameters similar but not quite the same as the best parameters (monte carlo)
    # Then, calculate the evidence, pr, and other results for each set of parameters
    alpha_opt, dmax_opt = opt_params
    mult = 3.0

    if not single_proc:
        if nprocs == 0:
            n_proc = max(multiprocessing.cpu_count() - 1, 1)
        else:
            n_proc = min(nprocs, multiprocessing.cpu_count())
        mp_pool = multiprocessing.Pool(processes=n_proc)
        mp_get_evidence = functools.partial(getEvidence, q=q, i=i, err=err, N=N)
    else:
        n_proc = nprocs

    ev_array = np.zeros(mc_runs)
    c_array = np.zeros(mc_runs)
    f_array = np.zeros((mc_runs, N + 1))
    r_array = np.zeros((mc_runs, N + 1))

    max_dmax = dmax_opt + 0.1 * dmax_opt * 0.5 * mult

    _, ref_r = makePriorDistribution(i[0], N, max_dmax, "sphere")

    run_mc = True

    while run_mc:
        alpha_array = (
            alpha_opt + 0.1 * alpha_opt * (np.random.random(mc_runs) - 0.5) * mult
        )
        dmax_array = (
            dmax_opt + 0.1 * dmax_opt * (np.random.random(mc_runs) - 0.5) * mult
        )
        alpha_array[0] = alpha_opt
        dmax_array[0] = dmax_opt

        pts = list(zip(alpha_array, dmax_array))

        if not single_proc:
            try:
                results = mp_pool.map(mp_get_evidence, pts)
            except Exception:
                mp_pool.close()
                mp_pool.join()
                raise
        else:
            results = [getEvidence(params, q, i, err, N) for params in pts]

        for res_idx, res in enumerate(results):
            dmax = dmax_array[res_idx]

            evidence, c, f, r = res

            interp = scipy.interpolate.interp1d(r, f, copy=False)

            f_interp = np.zeros_like(ref_r)
            f_interp[ref_r < dmax] = interp(ref_r[ref_r < dmax])

            ev_array[res_idx] = evidence
            c_array[res_idx] = c
            f_array[res_idx, :] = f_interp
            r_array[res_idx, :] = ref_r

        if np.abs(ev_array).max() >= 9e8:
            mult = mult / 2.0

            if mult < 0.001:
                run_mc = False

            if abort_check.is_set():
                run_mc = False

        else:
            run_mc = False

    if not single_proc:
        mp_pool.close()
        mp_pool.join()

    # Then, calculate the probability of each result as exp(evidence - evidence_max)**(1/minimum_chisq), normalized by the sum of all result probabilities

    ev_max = ev_array.max()
    prob = np.exp(ev_array - ev_max) ** (1.0 / c_array.min())
    prob = prob / prob.sum()

    # Then, calculate the average P(r) function as the weighted sum of the P(r) functions
    # Then, calculate the error in P(r) as the square root of the weighted sum of squares of the difference between the average result and the individual estimate
    p_avg = np.sum(f_array * prob[:, None], axis=0)
    err = np.sqrt(np.abs(np.sum((f_array - p_avg) ** 2 * prob[:, None], axis=0)))

    # Then, calculate structural results as weighted sum of each result
    alpha = np.sum(alpha_array * prob)
    dmax = np.sum(dmax_array * prob)
    c = np.sum(c_array * prob)
    evidence = np.sum(ev_array * prob)

    area = np.trapz(f_array, r_array, axis=1)
    area2 = np.trapz(f_array * r_array**2, r_array, axis=1)
    rg_array = np.sqrt(abs(area2 / (2.0 * area)))
    i0_array = area * 4 * np.pi
    rg = np.sum(rg_array * prob)
    i0 = np.sum(i0_array * prob)

    sd_alpha = np.sqrt(np.sum((alpha_array - alpha) ** 2 * prob))
    sd_dmax = np.sqrt(np.sum((dmax_array - dmax) ** 2 * prob))
    sd_c = np.sqrt(np.sum((c_array - c) ** 2 * prob))
    sd_ev = np.sqrt(np.sum((ev_array - evidence) ** 2 * prob))
    sd_rg = np.sqrt(np.sum((rg_array - rg) ** 2 * prob))
    sd_i0 = np.sqrt(np.sum((i0_array - i0) ** 2 * prob))

    # Should I also extrapolate to q=0? Might be good, though maybe not in this function
    # Should I report number of good parameters (ftot(nmax-12 in Hansen code, line 2247))
    # Should I report number of Shannon Channels? That's easy to calculate: q_range*dmax/pi

    return (
        ref_r,
        p_avg,
        err,
        (alpha, sd_alpha),
        (dmax, sd_dmax),
        (c, sd_c),
        (evidence, sd_ev),
        (rg, sd_rg),
        (i0, sd_i0),
    )


def make_fit(q, r, pr):
    qr = np.outer(q, r)
    sinc_qr = np.where(qr == 0, 1, np.sin(qr) / qr)
    i = 4 * np.pi * np.trapz(pr * sinc_qr, r, axis=1)

    return i


def doBift(
    q,
    i,
    err,
    filename,
    npts,
    alpha_min,
    alpha_max,
    alpha_n,
    dmax_min,
    dmax_max,
    dmax_n,
    mc_runs,
    queue=None,
    abort_check=threading.Event(),
    single_proc=False,
    nprocs=0,
):
    # Clean up data
    start_idx = 0

    for j in range(i.size):
        if i[j] == 0 or err[j] == 0:
            start_idx = j + 1
        else:
            break

    end_idx = i.size

    for j in range(i.size, 0, -1):
        if i[j - 1] == 0 or err[j - 1] == 0:
            end_idx = j - 1
        else:
            break

    q = q[start_idx:end_idx]
    i = i[start_idx:end_idx]
    err = err[start_idx:end_idx]

    i_zeros = np.argwhere(i == 0)
    err_zeros = np.argwhere(err == 0)

    both_zeros = np.intersect1d(i_zeros, err_zeros)

    q = np.delete(q, both_zeros)
    i = np.delete(i, both_zeros)
    err = np.delete(err, both_zeros)

    if npts > len(q) // 2:
        npts = len(q) // 2

    # Start by finding the optimal dmax and alpha via minimization of evidence

    alpha_min = np.log(alpha_min)
    alpha_max = np.log(alpha_max)

    alpha_points = np.linspace(
        alpha_min, alpha_max, alpha_n
    )  # alpha points are log(alpha) for better search
    dmax_points = np.linspace(dmax_min, dmax_max, dmax_n)

    all_posteriors = np.zeros((dmax_points.size, alpha_points.size))

    N = npts - 1

    if not single_proc:
        if nprocs == 0:
            n_proc = max(multiprocessing.cpu_count() - 1, 1)
        else:
            n_proc = min(nprocs, multiprocessing.cpu_count())
        mp_pool = multiprocessing.Pool(processes=n_proc)
        mp_get_evidence = functools.partial(getEvidence, q=q, i=i, err=err, N=N)
    else:
        n_proc = nprocs

    # Loop through a range of dmax and alpha to get a starting point for the minimization

    if abort_check.is_set():
        if queue is not None:
            queue.put({"canceled": True})

        if not single_proc:
            mp_pool.close()
            mp_pool.join()

        return None

    for d_idx, dmax in enumerate(dmax_points):
        pts = [(alpha, dmax) for alpha in alpha_points]

        if not single_proc:
            try:
                results = mp_pool.map(mp_get_evidence, pts)
            except Exception:
                mp_pool.close()
                mp_pool.join()
                raise
        else:
            results = [getEvidence(params, q, i, err, N) for params in pts]

        for res_idx, res in enumerate(results):
            all_posteriors[d_idx, res_idx] = res[0]

        if queue is not None:
            bift_status = {
                "alpha": pts[-1][0],
                "evidence": res[0],
                "chi": res[1],  # Actually chi squared
                "dmax": pts[-1][1],
                "spoint": (d_idx + 1) * (res_idx + 1),
                "tpoint": alpha_points.size * dmax_points.size,
            }

            queue.put({"update": bift_status})

        if abort_check.is_set():
            if queue is not None:
                queue.put({"canceled": True})

            if not single_proc:
                mp_pool.close()
                mp_pool.join()

            return None

    if not single_proc:
        mp_pool.close()
        mp_pool.join()

    if queue is not None:
        bift_status = {
            "alpha": pts[-1][0],
            "evidence": res[0],
            "chi": res[1],  # Actually chi squared
            "dmax": pts[-1][1],
            "spoint": alpha_points.size * dmax_points.size,
            "tpoint": alpha_points.size * dmax_points.size,
            "status": "Running minimization",
        }

        queue.put({"update": bift_status})

    min_idx = np.unravel_index(
        np.argmax(all_posteriors, axis=None), all_posteriors.shape
    )

    min_dmax = dmax_points[min_idx[0]]
    min_alpha = alpha_points[min_idx[1]]

    # Once a starting point is found, do an actual minimization to find the best alpha/dmax
    opt_res = scipy.optimize.minimize(
        getEvidenceOptimize, (min_alpha, min_dmax), (q, i, err, N), method="Powell"
    )

    if abort_check.is_set():
        if queue is not None:
            queue.put({"canceled": True})
        return None

    if opt_res.get("success"):
        alpha, dmax = opt_res.get("x")

        if dmax > 0:
            evidence, c, f, r = getEvidence((alpha, dmax), q, i, err, N)

            if queue is not None:
                bift_status = {
                    "alpha": alpha,
                    "evidence": evidence,
                    "chi": c,  # Actually chi squared
                    "dmax": dmax,
                    "spoint": alpha_points.size * dmax_points.size,
                    "tpoint": alpha_points.size * dmax_points.size,
                    "status": "Calculating Monte Carlo errors",
                }

                queue.put({"update": bift_status})

            pr = f

            area = np.trapz(pr, r)
            area2 = np.trapz(np.array(pr) * np.array(r) ** 2, r)

            rg = np.sqrt(abs(area2 / (2.0 * area)))
            i0 = area * 4 * np.pi

            fit = make_fit(q, r, pr)

            q_extrap = np.arange(0, q[1] - q[0], q[1])
            q_extrap = np.concatenate((q_extrap, q))

            fit_extrap = make_fit(q_extrap, r, pr)

            # Use a monte carlo method to estimate the errors in pr function, values found
            err_calc = calc_bift_errors(
                (alpha, dmax),
                q,
                i,
                err,
                N,
                mc_runs,
                abort_check=abort_check,
                single_proc=single_proc,
                nprocs=n_proc,
            )

            if abort_check.is_set():
                if queue is not None:
                    queue.put({"canceled": True})
                return None

            r_err, _, pr_err, a_res, d_res, c_res, ev_res, rg_res, i0_res = err_calc

            # NOTE: Unlike Hansen, we don't return the average pr function from the montecarlo
            # error estimate, but rather the best pr from the optimal dmax/alpha found above
            # This is consistent with the old RAW behavior. In the future this could change.

            rg_sd = rg_res[1]
            i0_sd = i0_res[1]
            alpha_sd = a_res[1]
            dmax_sd = d_res[1]
            c_sd = c_res[1]
            ev_sd = ev_res[1]

            interp = scipy.interpolate.interp1d(r_err, pr_err, copy=False)
            err_interp = interp(r)

            results = {
                "dmax": dmax,  # Dmax
                "dmaxer": dmax_sd,  # Uncertainty in Dmax
                "rg": rg,  # Real space Rg
                "rger": rg_sd,  # Real space rg error
                "i0": i0,  # Real space I0
                "i0er": i0_sd,  # Real space I0 error
                "chisq": c,  # Actual chi squared value
                "chisq_er": c_sd,  # Uncertainty in chi squared
                "alpha": alpha,  # log(Alpha) used for the IFT
                "alpha_er": alpha_sd,  # Uncertainty in log(alpha)
                "evidence": evidence,  # Evidence of solution
                "evidence_er": ev_sd,  # Uncertainty in evidence of solution
                "qmin": q[0],  # Minimum q
                "qmax": q[-1],  # Maximum q
                "algorithm": "BIFT",
            }

            iftm = SASM.IFTM(
                pr, r, err_interp, i, q, err, fit, results, fit_extrap, q_extrap
            )

        else:
            if queue is not None:
                queue.put({"failed": True})
            return None
    else:
        if queue is not None:
            queue.put({"failed": True})
        return None

    return iftm



# -*- coding: utf-8 -*-
"""SASExceptions

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1BADH17VoXYqdvpPNN0C15U0ycdGGKceQ
"""

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import object, range, map, zip
from io import open
'''
Created on Aug 2, 2010

@author: specuser

#******************************************************************************
# This file is part of RAW.
#
#    RAW is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    RAW is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with RAW.  If not, see <http://www.gnu.org/licenses/>.
#
#******************************************************************************
'''

# from __future__ import absolute_import, division, print_function, unicode_literals
# from builtins import object, range, map, zip
# from io import open

class ImageLoadError(Exception):
       def __init__(self, value):
           self.parameter = value
       def __str__(self):
           return repr(self.parameter)

class WrongImageFormat(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class MaskSizeError(Exception):
   def __init__(self, value):
       self.parameter = value
   def __str__(self):
       return repr(self.parameter)

class UnrecognizedDataFormat(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class DataNotCompatible(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class InvalidQrange(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class CenterNotFound(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class AbsScaleNormFailed(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class NormalizationError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderLoadError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderMaskLoadError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class NoATSASError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class HeaderSaveError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class GNOMError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class CorMapError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)

class EFAError(Exception):
    def __init__(self, value):
       self.parameter = value
    def __str__(self):
       return repr(self.parameter)